# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test the rounds of the skill."""
import logging  # noqa: F401
from types import MappingProxyType
from typing import Dict, FrozenSet, Optional

from packages.valory.skills.abstract_round_abci.base import AbciAppDB
from packages.valory.skills.abstract_round_abci.base import (
    BaseSynchronizedData as SynchronizedData,
)
from packages.valory.skills.abstract_round_abci.base import MAX_INT_256
from packages.valory.skills.oracle_deployment_abci.payloads import (
    RandomnessPayload,
    SelectKeeperPayload,
)
from packages.valory.skills.price_estimation_abci.payloads import (
    EstimatePayload,
    ObservationPayload,
    TransactionHashPayload,
)
from packages.valory.skills.price_estimation_abci.rounds import (
    CollectObservationRound,
    EstimateConsensusRound,
)
from packages.valory.skills.price_estimation_abci.rounds import (
    Event as PriceEstimationEvent,
)
from packages.valory.skills.price_estimation_abci.rounds import (
    SynchronizedData as PriceEstimationSynchronizedSata,
)
from packages.valory.skills.price_estimation_abci.rounds import TxHashRound
from packages.valory.skills.registration_abci.rounds import (
    BaseSynchronizedData as RegistrationSynchronizedSata,
)
from packages.valory.skills.transaction_settlement_abci.payloads import ValidatePayload
from packages.valory.skills.transaction_settlement_abci.rounds import (
    Event as TransactionSettlementEvent,
)
from packages.valory.skills.transaction_settlement_abci.rounds import (
    RandomnessTransactionSubmissionRound,
)
from packages.valory.skills.transaction_settlement_abci.rounds import (
    SynchronizedData as TransactionSettlementSynchronizedSata,
)

from tests.test_skills.test_abstract_round_abci.test_base_rounds import (
    BaseCollectDifferentUntilThresholdRoundTest,
    BaseCollectSameUntilThresholdRoundTest,
)


MAX_PARTICIPANTS: int = 4
RANDOMNESS: str = "d1c29dce46f979f9748210d24bce4eae8be91272f5ca1a6aea2832d3dd676f51"


def get_participants() -> FrozenSet[str]:
    """Participants"""
    return frozenset([f"agent_{i}" for i in range(MAX_PARTICIPANTS)])


def get_participant_to_randomness(
    participants: FrozenSet[str], round_id: int
) -> Dict[str, RandomnessPayload]:
    """participant_to_randomness"""
    return {
        participant: RandomnessPayload(
            sender=participant,
            round_id=round_id,
            randomness=RANDOMNESS,
        )
        for participant in participants
    }


def get_most_voted_randomness() -> str:
    """most_voted_randomness"""
    return RANDOMNESS


def get_participant_to_selection(
    participants: FrozenSet[str],
) -> Dict[str, SelectKeeperPayload]:
    """participant_to_selection"""
    return {
        participant: SelectKeeperPayload(sender=participant, keeper="keeper")
        for participant in participants
    }


def get_most_voted_keeper_address() -> str:
    """most_voted_keeper_address"""
    return "keeper"


def get_safe_contract_address() -> str:
    """safe_contract_address"""
    return "0x6f6ab56aca12"


def get_participant_to_votes(
    participants: FrozenSet[str], vote: Optional[bool] = True
) -> Dict[str, ValidatePayload]:
    """participant_to_votes"""
    return {
        participant: ValidatePayload(sender=participant, vote=vote)
        for participant in participants
    }


def get_participant_to_observations(
    participants: FrozenSet[str],
) -> Dict[str, ObservationPayload]:
    """participant_to_observations"""
    return {
        participant: ObservationPayload(sender=participant, observation=1.0)
        for participant in participants
    }


def get_participant_to_estimate(
    participants: FrozenSet[str],
) -> Dict[str, EstimatePayload]:
    """participant_to_estimate"""
    return {
        participant: EstimatePayload(sender=participant, estimate=1.0)
        for participant in participants
    }


def get_estimate() -> float:
    """Estimate"""
    return 1.0


def get_most_voted_estimate() -> float:
    """most_voted_estimate"""
    return 1.0


def get_participant_to_tx_hash(
    participants: FrozenSet[str], hash_: Optional[str] = "tx_hash"
) -> Dict[str, TransactionHashPayload]:
    """participant_to_tx_hash"""
    return {
        participant: TransactionHashPayload(sender=participant, tx_hash=hash_)
        for participant in participants
    }


def get_most_voted_tx_hash() -> str:
    """most_voted_tx_hash"""
    return "tx_hash"


class TestRandomnessTransactionSubmissionRound(BaseCollectSameUntilThresholdRoundTest):
    """Test RandomnessTransactionSubmissionRound."""

    _synchronized_data_class = TransactionSettlementSynchronizedSata
    _event_class = TransactionSettlementEvent

    def test_run(
        self,
    ) -> None:
        """Run tests."""

        test_round = RandomnessTransactionSubmissionRound(
            self.synchronized_data, self.consensus_params
        )
        self._complete_run(
            self._test_round(
                test_round=test_round,
                round_payloads=get_participant_to_randomness(self.participants, 1),
                synchronized_data_update_fn=lambda _synchronized_data, _test_round: _synchronized_data.update(
                    participant_to_randomness=MappingProxyType(
                        dict(get_participant_to_randomness(self.participants, 1))
                    )
                ),
                synchronized_data_attr_checks=[
                    lambda _synchronized_data: _synchronized_data.participant_to_randomness.keys()
                ],
                most_voted_payload=RANDOMNESS,
                exit_event=TransactionSettlementEvent.DONE,
            )
        )


class TestCollectObservationRound(BaseCollectDifferentUntilThresholdRoundTest):
    """Test CollectObservationRound."""

    _synchronized_data_class = PriceEstimationSynchronizedSata
    _event_class = PriceEstimationEvent

    def test_run_a(
        self,
    ) -> None:
        """Runs tests."""

        test_round = CollectObservationRound(
            synchronized_data=self.synchronized_data,
            consensus_params=self.consensus_params,
        )
        self._complete_run(
            self._test_round(
                test_round=test_round,
                round_payloads=get_participant_to_observations(self.participants),
                synchronized_data_update_fn=lambda _synchronized_data, _: _synchronized_data.update(
                    participant_to_observations=get_participant_to_observations(
                        self.participants
                    )
                ),
                synchronized_data_attr_checks=[
                    lambda _synchronized_data: _synchronized_data.participant_to_observations.keys()
                ],
                exit_event=self._event_class.DONE,
            )
        )

    def test_run_b(
        self,
    ) -> None:
        """Runs tests with one less observation."""

        test_round = CollectObservationRound(
            synchronized_data=self.synchronized_data,
            consensus_params=self.consensus_params,
        )

        self._complete_run(
            self._test_round(
                test_round=test_round,
                round_payloads=get_participant_to_observations(
                    frozenset(list(self.participants)[:-1])
                ),
                synchronized_data_update_fn=lambda _synchronized_data, _: _synchronized_data.update(
                    participant_to_observations=get_participant_to_observations(
                        frozenset(list(self.participants)[:-1])
                    )
                ),
                synchronized_data_attr_checks=[
                    lambda _synchronized_data: _synchronized_data.participant_to_observations.keys()
                ],
                exit_event=self._event_class.DONE,
            )
        )


class TestEstimateConsensusRound(BaseCollectSameUntilThresholdRoundTest):
    """Test EstimateConsensusRound."""

    _synchronized_data_class = PriceEstimationSynchronizedSata
    _event_class = PriceEstimationEvent

    def test_run(
        self,
    ) -> None:
        """Runs test."""

        test_round = EstimateConsensusRound(
            synchronized_data=self.synchronized_data,
            consensus_params=self.consensus_params,
        )
        self._complete_run(
            self._test_round(
                test_round=test_round,
                round_payloads=get_participant_to_estimate(self.participants),
                synchronized_data_update_fn=lambda _synchronized_data, _test_round: _synchronized_data.update(
                    participant_to_estimate=dict(
                        get_participant_to_estimate(self.participants)
                    ),
                    most_voted_estimate=_test_round.most_voted_payload,
                ),
                synchronized_data_attr_checks=[
                    lambda _synchronized_data: _synchronized_data.participant_to_estimate.keys()
                ],
                most_voted_payload=1.0,
                exit_event=self._event_class.DONE,
            )
        )


class TestTxHashRound(BaseCollectSameUntilThresholdRoundTest):
    """Test TxHashRound."""

    _synchronized_data_class = PriceEstimationSynchronizedSata
    _event_class = PriceEstimationEvent

    def test_run(
        self,
    ) -> None:
        """Runs test."""

        test_round = TxHashRound(
            synchronized_data=self.synchronized_data,
            consensus_params=self.consensus_params,
        )

        hash_ = "tx_hash"
        self._complete_run(
            self._test_round(
                test_round=test_round,
                round_payloads=get_participant_to_tx_hash(self.participants, hash_),
                synchronized_data_update_fn=lambda _synchronized_data, _test_round: _synchronized_data,
                synchronized_data_attr_checks=[],
                most_voted_payload=hash_,
                exit_event=self._event_class.DONE,
            )
        )

    def test_run_none(
        self,
    ) -> None:
        """Runs test."""

        test_round = TxHashRound(
            synchronized_data=self.synchronized_data,
            consensus_params=self.consensus_params,
        )

        hash_ = None
        self._complete_run(
            self._test_round(
                test_round=test_round,
                round_payloads=get_participant_to_tx_hash(self.participants, hash_),
                synchronized_data_update_fn=lambda _synchronized_data, _test_round: _synchronized_data,
                synchronized_data_attr_checks=[],
                most_voted_payload=hash_,
                exit_event=self._event_class.NONE,
            )
        )


def test_synchronized_datas() -> None:
    """Test SynchronizedData."""

    participants = get_participants()
    participant_to_randomness = get_participant_to_randomness(participants, 1)
    most_voted_randomness = get_most_voted_randomness()
    participant_to_selection = get_participant_to_selection(participants)
    most_voted_keeper_address = get_most_voted_keeper_address()
    safe_contract_address = get_safe_contract_address()
    oracle_contract_address = get_safe_contract_address()
    participant_to_votes = get_participant_to_votes(participants)
    most_voted_tx_hash = get_most_voted_tx_hash()
    participant_to_observations = get_participant_to_observations(participants)
    estimate = get_estimate()
    most_voted_estimate = get_most_voted_estimate()

    synchronized_data = SynchronizedData(
        AbciAppDB(
            setup_data=AbciAppDB.data_to_lists(
                dict(
                    participants=participants,
                    participant_to_randomness=participant_to_randomness,
                    most_voted_randomness=most_voted_randomness,
                    participant_to_selection=participant_to_selection,
                    most_voted_keeper_address=most_voted_keeper_address,
                    participant_to_votes=participant_to_votes,
                )
            ),
        )
    )

    actual_keeper_randomness = int(most_voted_randomness, base=16) / MAX_INT_256
    assert synchronized_data.keeper_randomness == actual_keeper_randomness
    assert synchronized_data.most_voted_randomness == most_voted_randomness
    assert synchronized_data.most_voted_keeper_address == most_voted_keeper_address
    assert synchronized_data.participant_to_votes == participant_to_votes

    synchronized_data____ = RegistrationSynchronizedSata(
        AbciAppDB(
            setup_data=AbciAppDB.data_to_lists(
                dict(
                    participants=participants,
                    participant_to_randomness=participant_to_randomness,
                    most_voted_randomness=most_voted_randomness,
                    participant_to_selection=participant_to_selection,
                    most_voted_keeper_address=most_voted_keeper_address,
                )
            ),
        )
    )

    assert (
        abs(synchronized_data____.keeper_randomness - actual_keeper_randomness) < 1e-10
    )  # avoid equality comparisons between floats
    assert synchronized_data____.most_voted_randomness == most_voted_randomness
    assert synchronized_data____.most_voted_keeper_address == most_voted_keeper_address

    synchronized_data______ = PriceEstimationSynchronizedSata(
        AbciAppDB(
            setup_data=AbciAppDB.data_to_lists(
                dict(
                    participants=participants,
                    participant_to_randomness=participant_to_randomness,
                    most_voted_randomness=most_voted_randomness,
                    participant_to_selection=participant_to_selection,
                    most_voted_keeper_address=most_voted_keeper_address,
                    safe_contract_address=safe_contract_address,
                    oracle_contract_address=oracle_contract_address,
                    most_voted_tx_hash=most_voted_tx_hash,
                    most_voted_estimate=most_voted_estimate,
                    participant_to_observations=participant_to_observations,
                )
            ),
        )
    )

    synchronized_data______.set_aggregator_method("median")

    assert synchronized_data______.keeper_randomness == actual_keeper_randomness
    assert synchronized_data______.most_voted_randomness == most_voted_randomness
    assert (
        synchronized_data______.most_voted_keeper_address == most_voted_keeper_address
    )
    assert synchronized_data______.safe_contract_address == safe_contract_address
    assert synchronized_data______.oracle_contract_address == oracle_contract_address
    assert synchronized_data______.most_voted_tx_hash == most_voted_tx_hash
    assert synchronized_data______.most_voted_estimate == most_voted_estimate
    assert (
        synchronized_data______.participant_to_observations
        == participant_to_observations
    )
    assert synchronized_data______.estimate == estimate



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Common utils for scripts."""
import logging
import re
import subprocess  # nosec
import sys
from io import StringIO
from pathlib import Path
from typing import Match, cast

from aea.configurations.base import ProtocolSpecification
from aea.configurations.loader import ConfigLoader


SPECIFICATION_REGEX = re.compile(r"(---\nname.*\.\.\.)", re.DOTALL)
PROTOCOL_SPECIFICATION_ID_IN_SPECIFICATION_REGEX = re.compile(
    "^protocol_specification_id: (.*)$", re.MULTILINE
)
PACKAGES_DIR = Path("packages")


def setup_logger(name: str) -> logging.Logger:
    """Set up the logger."""
    format_ = "[%(asctime)s][%(levelname)s] %(message)s"
    logging.basicConfig(format=format_)
    logger_ = logging.getLogger(name)
    logger_.setLevel(logging.INFO)
    return logger_


logger = setup_logger(__name__)


def enforce(condition: bool, message: str = "") -> None:
    """Custom assertion."""
    if not condition:
        raise AssertionError(message)


def check_working_tree_is_dirty() -> None:
    """Check if the current Git working tree is dirty."""
    print("Checking whether the Git working tree is dirty...")
    result = subprocess.check_output(["git", "diff", "--stat"])  # nosec
    if len(result) > 0:
        print("Git working tree is dirty:")
        print(result.decode("utf-8"))
        sys.exit(1)
    else:
        print("All good!")


def load_protocol_specification_from_string(
    specification_content: str,
) -> ProtocolSpecification:
    """Load a protocol specification from string."""
    file = StringIO(initial_value=specification_content)
    config_loader = ConfigLoader(
        "protocol-specification_schema.json", ProtocolSpecification
    )
    protocol_spec = config_loader.load_protocol_specification(file)
    return protocol_spec


def get_protocol_specification_from_readme(package_path: Path) -> str:
    """Get the protocol specification from the package README."""
    logger.info("Get protocol specification from README %s", package_path)
    readme = package_path / "README.md"
    readme_content = readme.read_text()
    enforce(
        "## Specification" in readme_content,
        f"Cannot find specification section in {package_path}",
    )

    search_result = SPECIFICATION_REGEX.search(readme_content)
    enforce(
        search_result is not None,
        f"Cannot find specification section in README of {package_path}",
    )
    specification_content = cast(Match, search_result).group(0)
    # just for validation of the parsed string
    load_protocol_specification_from_string(specification_content)
    return specification_content


def get_protocol_specification_id_from_specification(specification: str) -> str:
    """Get the protocol specification id from the protocol specification."""
    matches = PROTOCOL_SPECIFICATION_ID_IN_SPECIFICATION_REGEX.findall(specification)
    enforce(
        len(matches) == 1,
        f"Expected exactly one protocol specification id, found: {matches}",
    )
    spec_id = matches[0]
    return spec_id



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Check dependencies."""
import re
import shutil
import subprocess  # nosec
from itertools import islice
from typing import Iterable, List, Pattern, Tuple


ERROR_MESSAGE_TEMPLATE_BINARY_NOT_FOUND = (
    "'{command}' is required by the "
    "abci connection, but it is not installed, "
    "or it is not accessible from the system path."
)
ERROR_MESSAGE_TEMPLATE_VERSION_TOO_LOW = (
    "The installed version of '{command}' "
    "is too low: expected at least {lower_bound}; "
    "found {actual_version}."
)

# for the purposes of this script,
# a version is a tuple of integers: (major, minor, patch)
VERSION = Tuple[int, int, int]
MINIMUM_TENDERMINT_VERSION: VERSION = (0, 34, 11)


def nth(iterable: Iterable, index: int, default: int = 0) -> int:
    """Returns the item at position 'index' or a default value"""
    return next(islice(iterable, index, None), default)


def get_version(*args: int) -> VERSION:
    """
    Get the version from a list of arguments.

    Set to '0' if there are not enough arguments.

    :param args: positional arguments
    :return: the version
    """
    major = nth(args, 0, 0)
    minor = nth(args, 1, 0)
    patch = nth(args, 2, 0)
    return major, minor, patch


def version_to_string(version: VERSION) -> str:
    """
    Transform version to string.

    :param version: the version.
    :return: the string representation.
    """
    return ".".join(map(str, version))


def print_ok_message(
    binary_name: str, actual_version: VERSION, version_lower_bound: VERSION
) -> None:  # pragma: nocover
    """
    Print OK message.

    :param binary_name: the binary binary_name.
    :param actual_version: the actual version.
    :param version_lower_bound: the version lower bound.
    """
    print(
        f"check '{binary_name}'>={version_to_string(version_lower_bound)}, "
        f"found {version_to_string(actual_version)}"
    )


def check_binary(
    binary_name: str,
    args: List[str],
    version_regex: Pattern,
    version_lower_bound: VERSION,
    only_warning: bool = False,
) -> None:  # pragma: nocover
    """
    Check a binary is accessible from the terminal.

    It breaks down in:
    1) check if the binary is reachable from the system path;
    2) check that the version number is higher or equal than the minimum required version.

    :param binary_name: the name of the binary.
    :param args: the arguments to provide to the binary to retrieve the version.
    :param version_regex: the regex used to extract the version from the output.
    :param version_lower_bound: the minimum required version.
    :param only_warning: if True, don't raise error but print a warning message
    """
    path = shutil.which(binary_name)
    if not path:
        message = ERROR_MESSAGE_TEMPLATE_BINARY_NOT_FOUND.format(command=binary_name)
        if only_warning:
            print("Warning: ", message)
            return
        raise ValueError(message)

    version_getter_command = [binary_name, *args]
    stdout = subprocess.check_output(version_getter_command).decode("utf-8")  # nosec
    version_match = version_regex.search(stdout)
    if version_match is None:
        print(
            f"Warning: cannot parse '{binary_name}' version "
            f"from command: {version_getter_command}. stdout: {stdout}"
        )
        return
    actual_version: VERSION = get_version(*map(int, version_match.groups(default="0")))
    if actual_version < version_lower_bound:
        message = ERROR_MESSAGE_TEMPLATE_VERSION_TOO_LOW.format(
            command=binary_name,
            lower_bound=version_to_string(version_lower_bound),
            actual_version=version_to_string(actual_version),
        )
        if only_warning:
            print(f"Warning: {message}")
            return
        raise ValueError(message)

    print_ok_message(binary_name, actual_version, version_lower_bound)


def check_versions() -> None:  # pragma: nocover
    """Check versions."""
    check_binary(
        "tendermint",
        ["version"],
        re.compile(r"([0-9]+)\.([0-9]+)\.([0-9]+)"),
        MINIMUM_TENDERMINT_VERSION,
        only_warning=True,
    )


def main() -> None:  # pragma: nocover
    """The main entrypoint of the script."""
    check_versions()


if __name__ == "__main__":
    main()  # pragma: nocover



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

import os
import re
from typing import Dict

from setuptools import find_packages, setup  # type: ignore

PACKAGE_NAME = "autonomy"
here = os.path.abspath(os.path.dirname(__file__))


def get_all_extras() -> Dict:

    cli_deps = [
        "click==8.0.2",
    ]

    extras = {
        "cli": cli_deps,
    }

    # add "all" extras
    extras["all"] = list(set(dep for e in extras.values() for dep in e))
    return extras


all_extras = get_all_extras()

base_deps = [
    "Flask>=2.0.2,<3.0.0",
    "open-aea[all]>=1.10.0,<2.0.0",
    "pandas<1.4,>=1.3.4",
]
base_deps.extend(all_extras["cli"])

here = os.path.abspath(os.path.dirname(__file__))
about: Dict[str, str] = {}
with open(os.path.join(here, PACKAGE_NAME, "__version__.py"), "r") as f:
    exec(f.read(), about)


def parse_readme():
    with open("README.md", "r") as f:
        readme = f.read()

    # replace relative links of images
    raw_url_root = (
        "https://raw.githubusercontent.com/valory-xyz/open-autonomy/main/"
    )
    replacement = raw_url_root + r"\g<0>"
    readme = re.sub(r"(?<=<img src=\")(/.*)(?=\")", replacement, readme, re.DOTALL)

    return "\n".join([readme])


if __name__ == "__main__":
    setup(
        name=about["__title__"],
        description=about["__description__"],
        version=about["__version__"],
        author=about["__author__"],
        url=about["__url__"],
        long_description=parse_readme(),
        long_description_content_type="text/markdown",
        package_data={"autonomy": ["py.typed"]},
        packages=find_packages(include=["autonomy*"]),
        classifiers=[
            "Environment :: Console",
            "Environment :: Web Environment",
            "Development Status :: 2 - Pre-Alpha",
            "Intended Audience :: Developers",
            "License :: OSI Approved :: Apache Software License",
            "Natural Language :: English",
            "Operating System :: MacOS",
            "Operating System :: Microsoft",
            "Operating System :: Unix",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "Programming Language :: Python :: 3.10",
            "Topic :: Communications",
            "Topic :: Internet",
            "Topic :: Scientific/Engineering",
            "Topic :: Software Development",
            "Topic :: System",
        ],
        install_requires=base_deps,
        tests_require=["tox"],
        extras_require=all_extras,
        entry_points={"console_scripts": ["autonomy=autonomy.cli:cli"]},
        zip_safe=False,
        include_package_data=True,
        license=about["__license__"],
        python_requires=">=3.7",
        keywords="autonomy open-autonomy aea open-aea autonomous-economic-agents agent-framework multi-agent-systems multi-agent cryptocurrency cryptocurrencies dezentralized dezentralized-network",
        project_urls={
            "Bug Reports": "https://github.com/valory-xyz/open-autonomy/issues",
            "Source": "https://github.com/valory/open-autonomy",
        },
    )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021 valory
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains class representations corresponding to every custom type in the protocol specification."""
import datetime
from enum import Enum
from typing import List, Optional

from aea.exceptions import enforce

from packages.valory.protocols.abci import abci_pb2


class BlockParams:
    """This class represents an instance of BlockParams."""

    __slots__ = ["max_bytes", "max_gas"]

    def __init__(self, max_bytes: int, max_gas: int):
        """Initialise an instance of BlockParams."""
        self.max_bytes = max_bytes
        self.max_gas = max_gas

    @staticmethod
    def encode(
        block_params_protobuf_object, block_params_object: "BlockParams"
    ) -> None:
        """
        Encode an instance of this class into the protocol buffer object.

        The protocol buffer object in the block_params_protobuf_object argument is matched with the instance of this class in the 'block_params_object' argument.

        :param block_params_protobuf_object: the protocol buffer object whose type corresponds with this class.
        :param block_params_object: an instance of this class to be encoded in the protocol buffer object.
        """
        block_params_protobuf_object.max_bytes = block_params_object.max_bytes
        block_params_protobuf_object.max_gas = block_params_object.max_gas

    @classmethod
    def decode(cls, block_params_protobuf_object) -> "BlockParams":
        """
        Decode a protocol buffer object that corresponds with this class into an instance of this class.

        A new instance of this class is created that matches the protocol buffer object in the 'block_params_protobuf_object' argument.

        :param block_params_protobuf_object: the protocol buffer object whose type corresponds with this class.
        :return: A new instance of this class that matches the protocol buffer object in the 'block_params_protobuf_object' argument.
        """
        return BlockParams(
            block_params_protobuf_object.max_bytes, block_params_protobuf_object.max_gas
        )

    def __eq__(self, other) -> bool:
        """Compare with another object."""
        return (
            isinstance(other, BlockParams)
            and self.max_bytes == other.max_bytes
            and self.max_gas == other.max_gas
        )


class Duration:
    """This class represents an instance of Duration."""

    __slots__ = ["seconds", "nanos"]

    def __init__(self, seconds: int, nanos: int):
        """
        Initialise an instance of Duration.

        :param seconds: Signed seconds of the span of time.
            Must be from -315,576,000,000 to +315,576,000,000 inclusive.
        :param nanos: Signed fractions of a second at nanosecond resolution of the span of time.
            Durations less than one second are represented with a 0 seconds field and
            a positive or negative nanos field. For durations of one second or more,
            a non-zero value for the nanos field must be of the same sign as the seconds field.
            Must be from -999,999,999 to +999,999,999 inclusive.
        """
        self.seconds = seconds
        self.nanos = nanos

    @staticmethod
    def encode(duration_protobuf_object, duration_object: "Duration") -> None:
        """
        Encode an instance of this class into the protocol buffer object.

        The protocol buffer object in the duration_protobuf_object argument is matched with the instance of this class in the 'duration_object' argument.

        :param duration_protobuf_object: the protocol buffer object whose type corresponds with this class.
        :param duration_object: an instance of this class to be encoded in the protocol buffer object.
        """
        duration_protobuf_object.seconds = duration_object.seconds
        duration_protobuf_object.nanos = duration_object.nanos

    @classmethod
    def decode(cls, duration_protobuf_object) -> "Duration":
        """
        Decode a protocol buffer object that corresponds with this class into an instance of this class.

        A new instance of this class is created that matches the protocol buffer object in the 'duration_protobuf_object' argument.

        :param duration_protobuf_object: the protocol buffer object whose type corresponds with this class.
        :return: A new instance of this class that matches the protocol buffer object in the 'duration_protobuf_object' argument.
        """
        return Duration(
            duration_protobuf_object.seconds, duration_protobuf_object.nanos
        )

    def __eq__(self, other) -> bool:
        """Compare with another object."""
        return (
            isinstance(other, Duration)
            and self.seconds == other.seconds
            and self.nanos == other.nanos
        )


class EvidenceParams:
    """This class represents an instance of EvidenceParams."""

    __slots__ = [
        "max_age_num_blocks",
        "max_age_duration",
        "max_bytes",
    ]

    def __init__(
        self, max_age_num_blocks: int, max_age_duration: Duration, max_bytes: int
    ):
        """Initialise an instance of BlockParams."""
        self.max_age_num_blocks = max_age_num_blocks
        self.max_age_duration = max_age_duration
        self.max_bytes = max_bytes

    @staticmethod
    def encode(
        evidence_params_protobuf_object, evidence_params_object: "EvidenceParams"
    ) -> None:
        """
        Encode an instance of this class into the protocol buffer object.

        The protocol buffer object in the evidence_params_protobuf_object argument is matched with the instance of this class in the 'evidence_params_object' argument.

        :param evidence_params_protobuf_object: the protocol buffer object whose type corresponds with this class.
        :param evidence_params_object: an instance of this class to be encoded in the protocol buffer object.
        """
        evidence_params_protobuf_object.max_age_num_blocks = (
            evidence_params_object.max_age_num_blocks
        )
        Duration.encode(
            evidence_params_protobuf_object.max_age_duration,
            evidence_params_object.max_age_duration,
        )
        evidence_params_protobuf_object.max_bytes = evidence_params_object.max_bytes

    @classmethod
    def decode(cls, evidence_params_protobuf_object) -> "EvidenceParams":
        """
        Decode a protocol buffer object that corresponds with this class into an instance of this class.

        A new instance of this class is created that matches the protocol buffer object in the 'evidence_params_protobuf_object' argument.

        :param evidence_params_protobuf_object: the protocol buffer object whose type corresponds with this class.
        :return: A new instance of this class that matches the protocol buffer object in the 'evidence_params_protobuf_object' argument.
        """
        duration = Duration.decode(evidence_params_protobuf_object.max_age_duration)
        return EvidenceParams(
            evidence_params_protobuf_object.max_age_num_blocks,
            duration,
            evidence_params_protobuf_object.max_bytes,
        )

    def __eq__(self, other) -> bool:
        """Compare with another object."""
        return (
            isinstance(other, EvidenceParams)
            and self.max_age_num_blocks == other.max_age_num_blocks
            and self.max_age_duration == other.max_age_duration
            and self.max_bytes == other.max_bytes
        )


class ValidatorParams:
    """This class represents an instance of ValidatorParams."""

    __slots__ = [
        "pub_key_types",
    ]

    def __init__(self, pub_key_types: List[str]):
        """Initialise an instance of BlockParams."""
        self.pub_key_types = pub_key_types

    @staticmethod
    def encode(
        validator_params_protobuf_object, validator_params_object: "ValidatorParams"
    ) -> None:
        """
        Encode an instance of this class into the protocol buffer object.

        The protocol buffer object in the validator_params_protobuf_object argument is matched with the instance of this class in the 'validator_params_object' argument.

        :param validator_params_protobuf_object: the protocol buffer object whose type corresponds with this class.
        :param validator_params_object: an instance of this class to be encoded in the protocol buffer object.
        """
        validator_params_protobuf_object.pub_key_types.extend(
            validator_params_object.pub_key_types
        )

    @classmethod
    def decode(cls, validator_params_protobuf_object) -> "ValidatorParams":
        """
        Decode a protocol buffer object that corresponds with this class into an instance of this class.

        A new instance of this class is created that matches the protocol buffer object in the 'validator_params_protobuf_object' argument.

        :param validator_params_protobuf_object: the protocol buffer object whose type corresponds with this class.
        :return: A new instance of this class that matches the protocol buffer object in the 'validator_params_protobuf_object' argument.
        """
        pub_key_types = list(validator_params_protobuf_object.pub_key_types)
        return ValidatorParams(
            pub_key_types,
        )

    def __eq__(self, other) -> bool:
        """Compare with another object."""
        return (
            isinstance(other, ValidatorParams)
            and self.pub_key_types == other.pub_key_types
        )


class VersionParams:
    """This class represents an instance of VersionParams."""

    __slots__ = [
        "app_version",
    ]

    def __init__(self, app_version: int):
        """Initialise an instance of BlockParams."""
        self.app_version = app_version

    @staticmethod
    def encode(
        version_params_protobuf_object, version_params_object: "VersionParams"
    ) -> None:
        """
        Encode an instance of this class into the protocol buffer object.

        The protocol buffer object in the version_params_protobuf_object argument is matched with the instance of this class in the 'version_params_object' argument.

        :param version_params_protobuf_object: the protocol buffer object whose type corresponds with this class.
        :param version_params_object: an instance of this class to be encoded in the protocol buffer object.
        """
        version_params_protobuf_object.app_version = version_params_object.app_version

    @classmethod
    def decode(cls, version_params_protobuf_object) -> "VersionParams":
        """
        Decode a protocol buffer object that corresponds with this class into an instance of this class.

        A new instance of this class is created that matches the protocol buffer object in the 'version_params_protobuf_object' argument.

        :param version_params_protobuf_object: the protocol buffer object whose type corresponds with this class.
        :return: A new instance of this class that matches the protocol buffer object in the 'version_params_protobuf_object' argument.
        """
        app_version = version_params_protobuf_object.app_version
        return VersionParams(
            app_version,
        )

    def __eq__(self, other) -> bool:
        """Compare with another object."""
        return (
            isinstance(other, VersionParams) and self.app_version == other.app_version
        )


class ConsensusParams:
    """This class represents an instance of ConsensusParams."""

    def __init__(
        self,
        block: "BlockParams",
        evidence_params: "EvidenceParams",
        validator_params: "ValidatorParams",
        version_params: "VersionParams",
    ):
        """Initialise an instance of ConsensusParams."""
        self.block = block
        self.evidence_params = evidence_params
        self.validator_params = validator_params
        self.version_params = version_params

    @staticmethod
    def encode(
        consensus_params_protobuf_object,
        consensus_params_object: Optional["ConsensusParams"],
    ) -> None:
        """
        Encode an instance of this class into the protocol buffer object.

        The protocol buffer object in the consensus_params_protobuf_object argument is matched with the instance of this class in the 'consensus_params_object' argument.

        :param consensus_params_protobuf_object: the protocol buffer object whose type corresponds with this class.
        :param consensus_params_object: an instance of this class to be encoded in the protocol buffer object.
        """
        BlockParams.encode(
            consensus_params_protobuf_object.block, consensus_params_object.block
        )
        EvidenceParams.encode(
            consensus_params_protobuf_object.evidence,
            consensus_params_object.evidence_params,
        )
        ValidatorParams.encode(
            consensus_params_protobuf_object.validator,
            consensus_params_object.validator_params,
        )
        VersionParams.encode(
            consensus_params_protobuf_object.version,
            consensus_params_object.version_params,
        )

    @classmethod
    def decode(cls, consensus_params_protobuf_object) -> "ConsensusParams":
        """
        Decode a protocol buffer object that corresponds with this class into an instance of this class.

        A new instance of this class is created that matches the protocol buffer object in the 'consensus_params_protobuf_object' argument.

        :param consensus_params_protobuf_object: the protocol buffer object whose type corresponds with this class.
        :return: A new instance of this class that matches the protocol buffer object in the 'consensus_params_protobuf_object' argument.
        """
        block_params = BlockParams.decode(consensus_params_protobuf_object.block)
        evidence_params = EvidenceParams.decode(
            consensus_params_protobuf_object.evidence
        )
        validator_params = ValidatorParams.decode(
            consensus_params_protobuf_object.validator
        )
        version_params = VersionParams.decode(consensus_params_protobuf_object.version)
        return ConsensusParams(
            block_params, evidence_params, validator_params, version_params
        )

    def __eq__(self, other):
        """Compare with another object."""
        return (
            isinstance(other, ConsensusParams)
            and self.block == other.block
            and self.evidence_params == other.evidence_params
            and self.validator_params == other.validator_params
            and self.version_params == other.version_params
        )


class EventAttribute:
    """This class represents an instance of EventAttribute."""

    def __init__(self, key: bytes, value: bytes, index: bool):
        """Initialise an instance of EventAttribute."""
        self.key = key
        self.value = value
        self.index = index

    @staticmethod
    def encode(
        event_attribute_protobuf_object, event_attribute_object: "EventAttribute"
    ) -> None:
        """
        Encode an instance of this class into the protocol buffer object.

        The protocol buffer object in the event_attribute_protobuf_object argument is matched with the instance of this class in the 'event_attribute_object' argument.

        :param event_attribute_protobuf_object: the protocol buffer object whose type corresponds with this class.
        :param event_attribute_object: an instance of this class to be encoded in the protocol buffer object.
        """
        event_attribute_protobuf_object.key = event_attribute_object.key
        event_attribute_protobuf_object.value = event_attribute_object.value
        event_attribute_protobuf_object.index = event_attribute_object.index

    @classmethod
    def decode(cls, event_attribute_protobuf_object) -> "EventAttribute":
        """
        Decode a protocol buffer object that corresponds with this class into an instance of this class.

        A new instance of this class is created that matches the protocol buffer object in the 'event_attribute_protobuf_object' argument.

        :param event_attribute_protobuf_object: the protocol buffer object whose type corresponds with this class.
        :return: A new instance of this class that matches the protocol buffer object in the 'event_attribute_protobuf_object' argument.
        """
        return EventAttribute(
            event_attribute_protobuf_object.key,
            event_attribute_protobuf_object.value,
            event_attribute_protobuf_object.index,
        )

    def __eq__(self, other):
        """Compare with another object."""
        return (
            isinstance(other, EventAttribute)
            and self.key == other.key
            and self.value == other.value
            and self.index == other.index
        )


class Event:
    """This class represents an instance of Event."""

    def __init__(self, type_: str, attributes: List[EventAttribute]):
        """Initialise an instance of Event."""
        self.type_ = type_
        self.attributes = attributes

    @staticmethod
    def encode(event_protobuf_object, event_object: "Event") -> None:
        """
        Encode an instance of this class into the protocol buffer object.

        The protocol buffer object in the event_protobuf_object argument is matched with the instance of this class in the 'event_object' argument.

        :param event_protobuf_object: the protocol buffer object whose type corresponds with this class.
        :param event_object: an instance of this class to be encoded in the protocol buffer object.
        """
        event_protobuf_object.type = event_object.type_

        event_attribute_protobuf_objects = []
        for event_attribute_object in event_object.attributes:
            event_attribute_protobuf_object = (
                abci_pb2.AbciMessage.Events.EventAttribute()
            )
            EventAttribute.encode(
                event_attribute_protobuf_object, event_attribute_object
            )
            event_attribute_protobuf_objects.append(event_attribute_protobuf_object)

        event_protobuf_object.attributes.extend(event_attribute_protobuf_objects)

    @classmethod
    def decode(cls, event_protobuf_object) -> "Event":
        """
        Decode a protocol buffer object that corresponds with this class into an instance of this class.

        A new instance of this class is created that matches the protocol buffer object in the 'event_protobuf_object' argument.

        :param event_protobuf_object: the protocol buffer object whose type corresponds with this class.
        :return: A new instance of this class that matches the protocol buffer object in the 'event_protobuf_object' argument.
        """
        attributes = []
        for event_attribute_protobuf_object in list(event_protobuf_object.attributes):
            attribute = EventAttribute.decode(event_attribute_protobuf_object)
            attributes.append(attribute)
        return Event(event_protobuf_object.type, attributes)

    def __eq__(self, other):
        """Compare with another object."""
        return (
            isinstance(other, Event)
            and self.type_ == other.type_
            and self.attributes == other.attributes
        )


class Events:
    """This class represents an instance of Events."""

    def __init__(self, events: List[Event]):
        """Initialise an instance of Events."""
        self.events = events

    @staticmethod
    def encode(events_protobuf_object, events_object: "Events") -> None:
        """
        Encode an instance of this class into the protocol buffer object.

        The protocol buffer object in the events_protobuf_object argument is matched with the instance of this class in the 'events_object' argument.

        :param events_protobuf_object: the protocol buffer object whose type corresponds with this class.
        :param events_object: an instance of this class to be encoded in the protocol buffer object.
        """
        event_protobuf_objects = []
        for event_object in events_object.events:
            event_protobuf_object = abci_pb2.AbciMessage.Events.Event()
            Event.encode(event_protobuf_object, event_object)
            event_protobuf_objects.append(event_protobuf_object)
        events_protobuf_object.events.extend(event_protobuf_objects)

    @classmethod
    def decode(cls, events_protobuf_object) -> "Events":
        """
        Decode a protocol buffer object that corresponds with this class into an instance of this class.

        A new instance of this class is created that matches the protocol buffer object in the 'events_protobuf_object' argument.

        :param events_protobuf_object: the protocol buffer object whose type corresponds with this class.
        :return: A new instance of this class that matches the protocol buffer object in the 'events_protobuf_object' argument.
        """
        event_objects = []
        for event_protobuf_object in list(events_protobuf_object.events):
            event_object = Event.decode(event_protobuf_object)
            event_objects.append(event_object)
        return Events(event_objects)

    def __eq__(self, other):
        """Compare with another object."""
        return isinstance(other, Events) and self.events == other.events


class EvidenceType(Enum):
    """This class represent an instance of EvidenceType."""

    UNKNOWN = 0
    DUPLICATE_VOTE = 1
    LIGHT_CLIENT_ATTACK = 2


class Evidence:
    """This class represent an instance of Evidence."""

    def __init__(
        self,
        evidence_type: EvidenceType,
        validator: "Validator",
        height: int,
        time: "Timestamp",
        total_voting_power: int,
    ):
        """Initialise an instance of Evidences."""
        self.evidence_type = evidence_type
        self.validator = validator
        self.height = height
        self.time = time
        self.total_voting_power = total_voting_power

    @staticmethod
    def encode(evidence_protobuf_object, evidence_object: "Evidence") -> None:
        """
        Encode an instance of this class into the protocol buffer object.

        The protocol buffer object in the evidence_protobuf_object argument is matched with the instance of this class in the 'evidence_object' argument.

        :param evidence_protobuf_object: the protocol buffer object whose type corresponds with this class.
        :param evidence_object: an instance of this class to be encoded in the protocol buffer object.
        """
        evidence_protobuf_object.type = evidence_object.evidence_type.value

        validator_protobuf_object = abci_pb2.AbciMessage.LastCommitInfo.Validator()
        Validator.encode(validator_protobuf_object, evidence_object.validator)
        evidence_protobuf_object.validator.CopyFrom(validator_protobuf_object)

        evidence_protobuf_object.height = evidence_object.height

        timestamp_protobuf_object = abci_pb2.AbciMessage.Timestamp()
        Timestamp.encode(timestamp_protobuf_object, evidence_object.time)
        evidence_protobuf_object.time.CopyFrom(timestamp_protobuf_object)

        evidence_protobuf_object.total_voting_power = evidence_object.total_voting_power

    @classmethod
    def decode(cls, evidence_protobuf_object) -> "Evidence":
        """
        Decode a protocol buffer object that corresponds with this class into an instance of this class.

        A new instance of this class is created that matches the protocol buffer object in the 'evidence_protobuf_object' argument.

        :param evidence_protobuf_object: the protocol buffer object whose type corresponds with this class.
        :return: A new instance of this class that matches the protocol buffer object in the 'evidence_protobuf_object' argument.
        """
        evidence_type = EvidenceType(evidence_protobuf_object.type)
        validator = Validator.decode(evidence_protobuf_object.validator)
        height = evidence_protobuf_object.height
        time = Timestamp.decode(evidence_protobuf_object.time)
        total_voting_power = evidence_protobuf_object.total_voting_power
        return Evidence(evidence_type, validator, height, time, total_voting_power)

    def __eq__(self, other):
        """Compare with another object."""
        return (
            isinstance(other, Evidence)
            and self.evidence_type == other.evidence_type
            and self.validator == other.validator
            and self.height == other.height
            and self.time == other.time
            and self.total_voting_power == other.total_voting_power
        )


class Evidences:
    """This class represents an instance of Evidences."""

    def __init__(self, byzantine_validators: List[Evidence]):
        """Initialise an instance of Evidences."""
        self.byzantine_validators = byzantine_validators

    @staticmethod
    def encode(evidences_protobuf_object, evidences_object: "Evidences") -> None:
        """
        Encode an instance of this class into the protocol buffer object.

        The protocol buffer object in the evidences_protobuf_object argument is matched with the instance of this class in the 'evidences_object' argument.

        :param evidences_protobuf_object: the protocol buffer object whose type corresponds with this class.
        :param evidences_object: an instance of this class to be encoded in the protocol buffer object.
        """
        validators_protobuf_objects = []
        for validator in evidences_object.byzantine_validators:
            evidence_protobuf_object = abci_pb2.AbciMessage.Evidences.Evidence()
            Evidence.encode(evidence_protobuf_object, validator)
            validators_protobuf_objects.append(evidence_protobuf_object)
        evidences_protobuf_object.byzantine_validators.extend(
            validators_protobuf_objects
        )

    @classmethod
    def decode(cls, evidences_protobuf_object) -> "Evidences":
        """
        Decode a protocol buffer object that corresponds with this class into an instance of this class.

        A new instance of this class is created that matches the protocol buffer object in the 'evidences_protobuf_object' argument.

        :param evidences_protobuf_object: the protocol buffer object whose type corresponds with this class.
        :return: A new instance of this class that matches the protocol buffer object in the 'evidences_protobuf_object' argument.
        """
        validator_objects = []
        for validator_protobuf_object in list(
            evidences_protobuf_object.byzantine_validators
        ):
            validator_object = Evidence.decode(validator_protobuf_object)
            validator_objects.append(validator_object)
        return Evidences(validator_objects)

    def __eq__(self, other):
        """Compare with another object."""
        return (
            isinstance(other, Evidences)
            and self.byzantine_validators == other.byzantine_validators
        )


class CheckTxTypeEnum(Enum):
    """CheckTxTypeEnum for tx check."""

    NEW = 0
    RECHECK = 1


class CheckTxType:
    """This class represents an instance of CheckTxType."""

    def __init__(self, check_tx_type: CheckTxTypeEnum):
        """Initialise an instance of CheckTxType."""
        self.check_tx_type = check_tx_type

    @staticmethod
    def encode(
        check_tx_type_protobuf_object, check_tx_type_object: "CheckTxType"
    ) -> None:
        """
        Encode an instance of this class into the protocol buffer object.

        The protocol buffer object in the check_tx_type_protobuf_object argument is matched with the instance of this class in the 'check_tx_type_object' argument.

        :param check_tx_type_protobuf_object: the protocol buffer object whose type corresponds with this class.
        :param check_tx_type_object: an instance of this class to be encoded in the protocol buffer object.
        """
        check_tx_type_protobuf_object.type = check_tx_type_object.check_tx_type.value

    @classmethod
    def decode(cls, check_tx_type_protobuf_object) -> "CheckTxType":
        """
        Decode a protocol buffer object that corresponds with this class into an instance of this class.

        A new instance of this class is created that matches the protocol buffer object in the 'check_tx_type_protobuf_object' argument.

        :param check_tx_type_protobuf_object: the protocol buffer object whose type corresponds with this class.
        :return: A new instance of this class that matches the protocol buffer object in the 'check_tx_type_protobuf_object' argument.
        """
        return CheckTxType(CheckTxTypeEnum(check_tx_type_protobuf_object.type))

    def __eq__(self, other):
        """Compare with another object."""
        return (
            isinstance(other, CheckTxType) and self.check_tx_type == other.check_tx_type
        )


class ConsensusVersion:
    """This class represents an instance of ConsensusVersion."""

    def __init__(self, block: int, app: int):
        """Initialise an instance of ConsensusVersion."""
        self.block = block
        self.app = app

    @staticmethod
    def encode(
        consensus_version_protobuf_object, consensus_version_object: "ConsensusVersion"
    ) -> None:
        """
        Encode an instance of this class into the protocol buffer object.

        The protocol buffer object in the consensus_version_protobuf_object argument is matched with the instance of this class in the 'consensus_version_object' argument.

        :param consensus_version_protobuf_object: the protocol buffer object whose type corresponds with this class.
        :param consensus_version_object: an instance of this class to be encoded in the protocol buffer object.
        """
        consensus_version_protobuf_object.block = consensus_version_object.block
        consensus_version_protobuf_object.app = consensus_version_object.app

    @classmethod
    def decode(cls, consensus_version_protobuf_object) -> "ConsensusVersion":
        """
        Decode a protocol buffer object that corresponds with this class into an instance of this class.

        A new instance of this class is created that matches the protocol buffer object in the 'consensus_version_protobuf_object' argument.

        :param consensus_version_protobuf_object: the protocol buffer object whose type corresponds with this class.
        :return: A new instance of this class that matches the protocol buffer object in the 'consensus_version_protobuf_object' argument.
        """
        return ConsensusVersion(
            consensus_version_protobuf_object.block,
            consensus_version_protobuf_object.app,
        )

    def __eq__(self, other):
        """Compare with another object."""
        return (
            isinstance(other, ConsensusVersion)
            and self.block == other.block
            and self.app == other.app
        )


class PartSetHeader:
    """This class represents an instance of PartSetHeader."""

    def __init__(self, total: int, hash_: bytes):
        """Initialise an instance of PartSetHeader."""
        self.total = total
        self.hash_ = hash_

    @staticmethod
    def encode(
        part_set_header_protobuf_object, part_set_header_object: "PartSetHeader"
    ) -> None:
        """
        Encode an instance of this class into the protocol buffer object.

        The protocol buffer object in the part_set_header_protobuf_object argument is matched with the instance of this class in the 'part_set_header_object' argument.

        :param part_set_header_protobuf_object: the protocol buffer object whose type corresponds with this class.
        :param part_set_header_object: an instance of this class to be encoded in the protocol buffer object.
        """
        part_set_header_protobuf_object.total = part_set_header_object.total
        part_set_header_protobuf_object.hash = part_set_header_object.hash_

    @classmethod
    def decode(cls, part_set_header_protobuf_object) -> "PartSetHeader":
        """
        Decode a protocol buffer object that corresponds with this class into an instance of this class.

        A new instance of this class is created that matches the protocol buffer object in the 'part_set_header_protobuf_object' argument.

        :param part_set_header_protobuf_object: the protocol buffer object whose type corresponds with this class.
        :return: A new instance of this class that matches the protocol buffer object in the 'part_set_header_protobuf_object' argument.
        """
        return PartSetHeader(
            part_set_header_protobuf_object.total,
            part_set_header_protobuf_object.hash,
        )

    def __eq__(self, other):
        """Compare with another object."""
        return (
            isinstance(other, PartSetHeader)
            and self.total == other.total
            and self.hash_ == other.hash_
        )


class BlockID:
    """This class represents an instance of BlockID."""

    def __init__(self, hash_: bytes, part_set_header: PartSetHeader):
        """Initialise an instance of BlockID."""
        self.hash_ = hash_
        self.part_set_header = part_set_header

    @staticmethod
    def encode(block_id_protobuf_object, block_id_object: "BlockID") -> None:
        """
        Encode an instance of this class into the protocol buffer object.

        The protocol buffer object in the block_id_protobuf_object argument is matched with the instance of this class in the 'block_id_object' argument.

        :param block_id_protobuf_object: the protocol buffer object whose type corresponds with this class.
        :param block_id_object: an instance of this class to be encoded in the protocol buffer object.
        """
        block_id_protobuf_object.hash = block_id_object.hash_
        part_set_header_protobuf_object = abci_pb2.AbciMessage.Header.PartSetHeader()
        PartSetHeader.encode(
            part_set_header_protobuf_object, block_id_object.part_set_header
        )
        block_id_protobuf_object.part_set_header.CopyFrom(
            part_set_header_protobuf_object
        )

    @classmethod
    def decode(cls, block_id_protobuf_object) -> "BlockID":
        """
        Decode a protocol buffer object that corresponds with this class into an instance of this class.

        A new instance of this class is created that matches the protocol buffer object in the 'block_id_protobuf_object' argument.

        :param block_id_protobuf_object: the protocol buffer object whose type corresponds with this class.
        :return: A new instance of this class that matches the protocol buffer object in the 'block_id_protobuf_object' argument.
        """
        part_set_header = PartSetHeader.decode(block_id_protobuf_object.part_set_header)
        return BlockID(
            block_id_protobuf_object.hash,
            part_set_header,
        )

    def __eq__(self, other):
        """Compare with another object."""
        return (
            isinstance(other, BlockID)
            and self.hash_ == other.hash_
            and self.part_set_header == other.part_set_header
        )


class Header:  # pylint: disable=too-many-instance-attributes
    """This class represents an instance of Header."""

    def __init__(  # pylint: disable=too-many-arguments
        self,
        version: ConsensusVersion,
        chain_id: str,
        height: int,
        time: "Timestamp",
        last_block_id: BlockID,
        last_commit_hash: bytes,
        data_hash: bytes,
        validators_hash: bytes,
        next_validators_hash: bytes,
        consensus_hash: bytes,
        app_hash: bytes,
        last_results_hash: bytes,
        evidence_hash: bytes,
        proposer_address: bytes,
    ):
        """Initialise an instance of Header."""
        self.version = version
        self.chain_id = chain_id
        self.height = height
        self.time = time
        self.last_block_id = last_block_id
        self.last_commit_hash = last_commit_hash
        self.data_hash = data_hash
        self.validators_hash = validators_hash
        self.next_validators_hash = next_validators_hash
        self.consensus_hash = consensus_hash
        self.app_hash = app_hash
        self.last_results_hash = last_results_hash
        self.evidence_hash = evidence_hash
        self.proposer_address = proposer_address

    @property
    def timestamp(self) -> datetime.datetime:
        """Get the block timestamp."""
        timestamp: Timestamp = self.time
        nanoseconds = timestamp.nanos / 10 ** 9
        seconds = timestamp.seconds
        return datetime.datetime.fromtimestamp(seconds + nanoseconds)

    @staticmethod
    def encode(header_protobuf_object, header_object: "Header") -> None:
        """
        Encode an instance of this class into the protocol buffer object.

        The protocol buffer object in the header_protobuf_object argument is matched with the instance of this class in the 'header_object' argument.

        :param header_protobuf_object: the protocol buffer object whose type corresponds with this class.
        :param header_object: an instance of this class to be encoded in the protocol buffer object.
        """
        consensus_version_protobuf_obj = abci_pb2.AbciMessage.Header.ConsensusVersion()
        ConsensusVersion.encode(consensus_version_protobuf_obj, header_object.version)
        header_protobuf_object.version.CopyFrom(consensus_version_protobuf_obj)

        header_protobuf_object.chain_id = header_object.chain_id
        header_protobuf_object.height = header_object.height

        timestamp_protobuf_obj = abci_pb2.AbciMessage.Timestamp()
        Timestamp.encode(timestamp_protobuf_obj, header_object.time)
        header_protobuf_object.time.CopyFrom(timestamp_protobuf_obj)

        last_block_id_protobuf_obj = abci_pb2.AbciMessage.Header.BlockID()
        BlockID.encode(last_block_id_protobuf_obj, header_object.last_block_id)
        header_protobuf_object.last_block_id.CopyFrom(last_block_id_protobuf_obj)

        header_protobuf_object.last_commit_hash = header_object.last_commit_hash
        header_protobuf_object.data_hash = header_object.data_hash
        header_protobuf_object.validators_hash = header_object.validators_hash
        header_protobuf_object.next_validators_hash = header_object.next_validators_hash
        header_protobuf_object.consensus_hash = header_object.consensus_hash
        header_protobuf_object.app_hash = header_object.app_hash
        header_protobuf_object.last_results_hash = header_object.last_results_hash
        header_protobuf_object.evidence_hash = header_object.evidence_hash
        header_protobuf_object.proposer_address = header_object.proposer_address

    @classmethod
    def decode(  # pylint: disable=too-many-locals
        cls, header_protobuf_object
    ) -> "Header":
        """
        Decode a protocol buffer object that corresponds with this class into an instance of this class.

        A new instance of this class is created that matches the protocol buffer object in the 'header_protobuf_object' argument.

        :param header_protobuf_object: the protocol buffer object whose type corresponds with this class.
        :return: A new instance of this class that matches the protocol buffer object in the 'header_protobuf_object' argument.
        """
        consensus_version_obj = ConsensusVersion.decode(header_protobuf_object.version)
        chain_id = header_protobuf_object.chain_id
        height = header_protobuf_object.height
        time = Timestamp.decode(header_protobuf_object.time)

        last_block_id = BlockID.decode(header_protobuf_object.last_block_id)

        last_commit_hash = header_protobuf_object.last_commit_hash
        data_hash = header_protobuf_object.data_hash
        validators_hash = header_protobuf_object.validators_hash
        next_validators_hash = header_protobuf_object.next_validators_hash
        consensus_hash = header_protobuf_object.consensus_hash
        app_hash = header_protobuf_object.app_hash
        last_results_hash = header_protobuf_object.last_results_hash
        evidence_hash = header_protobuf_object.evidence_hash
        proposer_address = header_protobuf_object.proposer_address
        return Header(
            consensus_version_obj,
            chain_id,
            height,
            time,
            last_block_id,
            last_commit_hash,
            data_hash,
            validators_hash,
            next_validators_hash,
            consensus_hash,
            app_hash,
            last_results_hash,
            evidence_hash,
            proposer_address,
        )

    def __eq__(self, other):
        """Compare with another object."""
        return (
            isinstance(other, Header)
            and self.version == other.version
            and self.chain_id == other.chain_id
            and self.height == other.height
            and self.time == other.time
            and self.last_block_id == other.last_block_id
            and self.last_commit_hash == other.last_commit_hash
            and self.data_hash == other.data_hash
            and self.validators_hash == other.validators_hash
            and self.next_validators_hash == other.next_validators_hash
            and self.consensus_hash == other.consensus_hash
            and self.app_hash == other.app_hash
            and self.last_results_hash == other.last_results_hash
            and self.evidence_hash == other.evidence_hash
            and self.proposer_address == other.proposer_address
        )


class Validator:
    """This class represents an instance of Validator."""

    def __init__(self, address: bytes, power: int):
        """Initialise an instance of Validator."""
        self.address = address
        self.power = power

    @staticmethod
    def encode(validator_protobuf_object, validator_object: "Validator") -> None:
        """
        Encode an instance of this class into the protocol buffer object.

        The protocol buffer object in the validator_protobuf_object argument is matched with the instance of this class in the 'validator_object' argument.

        :param validator_protobuf_object: the protocol buffer object whose type corresponds with this class.
        :param validator_object: an instance of this class to be encoded in the protocol buffer object.
        """
        validator_protobuf_object.address = validator_object.address
        validator_protobuf_object.power = validator_object.power

    @classmethod
    def decode(cls, validator_protobuf_object) -> "Validator":
        """
        Decode a protocol buffer object that corresponds with this class into an instance of this class.

        A new instance of this class is created that matches the protocol buffer object in the 'validator_protobuf_object' argument.

        :param validator_protobuf_object: the protocol buffer object whose type corresponds with this class.
        :return: A new instance of this class that matches the protocol buffer object in the 'validator_protobuf_object' argument.
        """
        return Validator(
            validator_protobuf_object.address,
            validator_protobuf_object.power,
        )

    def __eq__(self, other):
        """Compare with another object."""
        return (
            isinstance(other, Validator)
            and self.address == other.address
            and self.power == other.power
        )


class VoteInfo:
    """This class represents an instance of VoteInfo."""

    def __init__(self, validator: Validator, signed_last_block: bool):
        """Initialise an instance of Validator."""
        self.validator = validator
        self.signed_last_block = signed_last_block

    @staticmethod
    def encode(vote_info_protobuf_object, vote_info_object: "VoteInfo") -> None:
        """
        Encode an instance of this class into the protocol buffer object.

        The protocol buffer object in the vote_info_protobuf_object argument is matched with the instance of this class in the 'vote_info_object' argument.

        :param vote_info_protobuf_object: the protocol buffer object whose type corresponds with this class.
        :param vote_info_object: an instance of this class to be encoded in the protocol buffer object.
        """
        validator_protobuf_object = abci_pb2.AbciMessage.LastCommitInfo.Validator()
        Validator.encode(validator_protobuf_object, vote_info_object.validator)
        vote_info_protobuf_object.validator.CopyFrom(validator_protobuf_object)

        vote_info_protobuf_object.signed_last_block = vote_info_object.signed_last_block

    @classmethod
    def decode(cls, vote_info_protobuf_object) -> "VoteInfo":
        """
        Decode a protocol buffer object that corresponds with this class into an instance of this class.

        A new instance of this class is created that matches the protocol buffer object in the 'vote_info_protobuf_object' argument.

        :param vote_info_protobuf_object: the protocol buffer object whose type corresponds with this class.
        :return: A new instance of this class that matches the protocol buffer object in the 'vote_info_protobuf_object' argument.
        """
        validator = Validator.decode(vote_info_protobuf_object.validator)
        return VoteInfo(
            validator,
            vote_info_protobuf_object.signed_last_block,
        )

    def __eq__(self, other):
        """Compare with another object."""
        return (
            isinstance(other, VoteInfo)
            and self.validator == other.validator
            and self.signed_last_block == other.signed_last_block
        )


class LastCommitInfo:
    """This class represents an instance of LastCommitInfo."""

    def __init__(self, round_: int, votes: List[VoteInfo]):
        """Initialise an instance of LastCommitInfo."""
        self.round_ = round_
        self.votes = votes

    @staticmethod
    def encode(
        last_commit_info_protobuf_object, last_commit_info_object: "LastCommitInfo"
    ) -> None:
        """
        Encode an instance of this class into the protocol buffer object.

        The protocol buffer object in the last_commit_info_protobuf_object argument is matched with the instance of this class in the 'last_commit_info_object' argument.

        :param last_commit_info_protobuf_object: the protocol buffer object whose type corresponds with this class.
        :param last_commit_info_object: an instance of this class to be encoded in the protocol buffer object.
        """
        last_commit_info_protobuf_object.round = last_commit_info_object.round_

        votes_protobuf_objects = []
        for vote_info in last_commit_info_object.votes:
            vote_info_protobuf_object = abci_pb2.AbciMessage.LastCommitInfo.VoteInfo()
            VoteInfo.encode(vote_info_protobuf_object, vote_info)
            votes_protobuf_objects.append(vote_info_protobuf_object)
        last_commit_info_protobuf_object.votes.extend(votes_protobuf_objects)

    @classmethod
    def decode(cls, last_commit_info_protobuf_object) -> "LastCommitInfo":
        """
        Decode a protocol buffer object that corresponds with this class into an instance of this class.

        A new instance of this class is created that matches the protocol buffer object in the 'last_commit_info_protobuf_object' argument.

        :param last_commit_info_protobuf_object: the protocol buffer object whose type corresponds with this class.
        :return: A new instance of this class that matches the protocol buffer object in the 'last_commit_info_protobuf_object' argument.
        """
        vote_info_objects = []
        for vote_info_protobuf_object in list(last_commit_info_protobuf_object.votes):
            vote_info = VoteInfo.decode(vote_info_protobuf_object)
            vote_info_objects.append(vote_info)

        return LastCommitInfo(last_commit_info_protobuf_object.round, vote_info_objects)

    def __eq__(self, other):
        """Compare with another object."""
        return (
            isinstance(other, LastCommitInfo)
            and self.round_ == other.round_
            and self.votes == other.votes
        )


class ProofOp:
    """This class represents an instance of ProofOp."""

    def __init__(self, type_: str, key: bytes, data: bytes):
        """
        Initialise an instance of ProofOp.

        ProofOp defines an operation used for calculating Merkle root
        The data could be arbitrary format, providing necessary data
        for example neighbouring node hash.

        :param type_: the type
        :param key: the key
        :param data: the data
        """
        self.type_ = type_
        self.key = key
        self.data = data

    @staticmethod
    def encode(proof_op_protobuf_object, proof_op_object: "ProofOp") -> None:
        """
        Encode an instance of this class into the protocol buffer object.

        The protocol buffer object in the proof_op_protobuf_object argument is matched with the instance of this class in the 'proof_op_object' argument.

        :param proof_op_protobuf_object: the protocol buffer object whose type corresponds with this class.
        :param proof_op_object: an instance of this class to be encoded in the protocol buffer object.
        """
        proof_op_protobuf_object.type = proof_op_object.type_
        proof_op_protobuf_object.key = proof_op_object.key
        proof_op_protobuf_object.data = proof_op_object.data

    @classmethod
    def decode(cls, proof_op_protobuf_object) -> "ProofOp":
        """
        Decode a protocol buffer object that corresponds with this class into an instance of this class.

        A new instance of this class is created that matches the protocol buffer object in the 'proof_op_protobuf_object' argument.

        :param proof_op_protobuf_object: the protocol buffer object whose type corresponds with this class.
        :return: A new instance of this class that matches the protocol buffer object in the 'proof_op_protobuf_object' argument.
        """
        return ProofOp(
            proof_op_protobuf_object.type,
            proof_op_protobuf_object.key,
            proof_op_protobuf_object.data,
        )

    def __eq__(self, other):
        """Compare with another object."""
        return (
            isinstance(other, ProofOp)
            and self.type_ == other.type_
            and self.key == other.key
            and self.data == other.data
        )


class ProofOps:
    """This class represents an instance of ProofOps."""

    def __init__(self, proof_ops: List[ProofOp]):
        """
        Initialise an instance of ProofOps.

        :param proof_ops: a list of ProofOp instances.
        """
        self.proof_ops = proof_ops

    @staticmethod
    def encode(proof_ops_protobuf_object, proof_ops_object: "ProofOps") -> None:
        """
        Encode an instance of this class into the protocol buffer object.

        The protocol buffer object in the proof_ops_protobuf_object argument is matched with the instance of this class in the 'proof_ops_object' argument.

        :param proof_ops_protobuf_object: the protocol buffer object whose type corresponds with this class.
        :param proof_ops_object: an instance of this class to be encoded in the protocol buffer object.
        """
        proof_ops_protobuf_objects = []
        for proof_op in proof_ops_object.proof_ops:
            proof_op_protobuf_object = abci_pb2.AbciMessage.ProofOps.ProofOp()
            ProofOp.encode(proof_op_protobuf_object, proof_op)
            proof_ops_protobuf_objects.append(proof_op_protobuf_object)
        proof_ops_protobuf_object.ops.extend(proof_ops_protobuf_objects)

    @classmethod
    def decode(cls, proof_ops_protobuf_object) -> "ProofOps":
        """
        Decode a protocol buffer object that corresponds with this class into an instance of this class.

        A new instance of this class is created that matches the protocol buffer object in the 'proof_ops_protobuf_object' argument.

        :param proof_ops_protobuf_object: the protocol buffer object whose type corresponds with this class.
        :return: A new instance of this class that matches the protocol buffer object in the 'proof_ops_protobuf_object' argument.
        """
        proof_ops_objects = []
        for proof_op_protobuf_object in list(proof_ops_protobuf_object.ops):
            proof_ops_objects.append(ProofOp.decode(proof_op_protobuf_object))
        return ProofOps(proof_ops_objects)

    def __eq__(self, other):
        """Compare with another object."""
        return isinstance(other, ProofOps) and self.proof_ops == other.proof_ops


class ResultType(Enum):
    """This class represents an instance of ResultType."""

    UNKNOWN = 0  # Unknown result, abort all snapshot restoration
    ACCEPT = 1  # Snapshot accepted, apply chunks
    ABORT = 2  # Abort all snapshot restoration
    REJECT = 3  # Reject this specific snapshot, try others
    REJECT_FORMAT = 4  # Reject all snapshots of this format, try others
    REJECT_SENDER = 5  # Reject all snapshots from the sender(s), try others


class Result:
    """This class represents an instance of Result."""

    def __init__(self, result_type: ResultType):
        """Initialise an instance of Result."""
        self.result_type = result_type

    @staticmethod
    def encode(result_protobuf_object, result_object: "Result") -> None:
        """
        Encode an instance of this class into the protocol buffer object.

        The protocol buffer object in the result_protobuf_object argument is matched with the instance of this class in the 'result_object' argument.

        :param result_protobuf_object: the protocol buffer object whose type corresponds with this class.
        :param result_object: an instance of this class to be encoded in the protocol buffer object.
        """
        result_protobuf_object.result_type = result_object.result_type.value

    @classmethod
    def decode(cls, result_protobuf_object) -> "Result":
        """
        Decode a protocol buffer object that corresponds with this class into an instance of this class.

        A new instance of this class is created that matches the protocol buffer object in the 'result_protobuf_object' argument.

        :param result_protobuf_object: the protocol buffer object whose type corresponds with this class.
        :return: A new instance of this class that matches the protocol buffer object in the 'result_protobuf_object' argument.
        """
        return Result(ResultType(result_protobuf_object.result_type))

    def __eq__(self, other):
        """Compare with another object."""
        return isinstance(other, Result) and self.result_type == other.result_type


class Snapshot:
    """This class represents an instance of Snapshot."""

    def __init__(
        self, height: int, format_: int, chunks: int, hash_: bytes, metadata: bytes
    ):
        """Initialise an instance of Snapshot."""
        self.height = height
        self.format_ = format_
        self.chunks = chunks
        self.hash_ = hash_
        self.metadata = metadata

    @staticmethod
    def encode(snapshot_protobuf_object, snapshot_object: "Snapshot") -> None:
        """
        Encode an instance of this class into the protocol buffer object.

        The protocol buffer object in the snapshot_protobuf_object argument is matched with the instance of this class in the 'snapshot_object' argument.

        :param snapshot_protobuf_object: the protocol buffer object whose type corresponds with this class.
        :param snapshot_object: an instance of this class to be encoded in the protocol buffer object.
        """
        snapshot_protobuf_object.height = snapshot_object.height
        snapshot_protobuf_object.format = snapshot_object.format_
        snapshot_protobuf_object.chunks = snapshot_object.chunks
        snapshot_protobuf_object.hash = snapshot_object.hash_
        snapshot_protobuf_object.metadata = snapshot_object.metadata

    @classmethod
    def decode(cls, snapshot_protobuf_object) -> "Snapshot":
        """
        Decode a protocol buffer object that corresponds with this class into an instance of this class.

        A new instance of this class is created that matches the protocol buffer object in the 'snapshot_protobuf_object' argument.

        :param snapshot_protobuf_object: the protocol buffer object whose type corresponds with this class.
        :return: A new instance of this class that matches the protocol buffer object in the 'snapshot_protobuf_object' argument.
        """
        return Snapshot(
            snapshot_protobuf_object.height,
            snapshot_protobuf_object.format,
            snapshot_protobuf_object.chunks,
            snapshot_protobuf_object.hash,
            snapshot_protobuf_object.metadata,
        )

    def __eq__(self, other):
        """Compare with another object."""
        return (
            isinstance(other, Snapshot)
            and self.height == other.height
            and self.format_ == other.format_
            and self.chunks == other.chunks
            and self.hash_ == other.hash_
            and self.metadata == other.metadata
        )


class SnapShots:
    """This class represents an instance of SnapShots."""

    def __init__(self, snapshots: List[Snapshot]):
        """Initialise an instance of SnapShots."""
        self.snapshots = snapshots

    @staticmethod
    def encode(snapshots_protobuf_object, snapshots_object: "SnapShots") -> None:
        """
        Encode an instance of this class into the protocol buffer object.

        The protocol buffer object in the snapshots_protobuf_object argument is matched with the instance of this class in the 'snapshots_object' argument.

        :param snapshots_protobuf_object: the protocol buffer object whose type corresponds with this class.
        :param snapshots_object: an instance of this class to be encoded in the protocol buffer object.
        """
        snapshot_protobuf_objects = []
        for snapshot_object in snapshots_object.snapshots:
            snapshot_protobuf_object = abci_pb2.AbciMessage.Snapshot()
            Snapshot.encode(snapshot_protobuf_object, snapshot_object)
            snapshot_protobuf_objects.append(snapshot_protobuf_object)
        snapshots_protobuf_object.snapshots.extend(snapshot_protobuf_objects)

    @classmethod
    def decode(cls, snapshots_protobuf_object) -> "SnapShots":
        """
        Decode a protocol buffer object that corresponds with this class into an instance of this class.

        A new instance of this class is created that matches the protocol buffer object in the 'snapshots_protobuf_object' argument.

        :param snapshots_protobuf_object: the protocol buffer object whose type corresponds with this class.
        :return: A new instance of this class that matches the protocol buffer object in the 'snapshots_protobuf_object' argument.
        """
        snapshot_objects = []
        for snapshot_protobuf_object in list(snapshots_protobuf_object.snapshots):
            snapshot_objects.append(Snapshot.decode(snapshot_protobuf_object))
        return SnapShots(snapshot_objects)

    def __eq__(self, other):
        """Compare with another object."""
        return isinstance(other, SnapShots) and self.snapshots == other.snapshots


class Timestamp:
    """This class represents an instance of Timestamp."""

    __slots__ = ["seconds", "nanos"]

    def __init__(self, seconds: int, nanos: int):
        """
        Initialise an instance of Timestamp.

        :param seconds: Represents seconds of UTC time since Unix epoch
            1970-01-01T00:00:00Z. Must be from 0001-01-01T00:00:00Z to
            9999-12-31T23:59:59Z inclusive.
        :param nanos: Non-negative fractions of a second at nanosecond resolution.
            Negative second values with fractions must still have non-negative nanos values
            that count forward in time. Must be from 0 to 999,999,999 inclusive.
        """
        self.seconds = seconds
        self.nanos = nanos
        enforce(
            0 <= nanos < 10 ** 9,
            "nanos argument must be from 0 to 999,999,999 inclusive",
            exception_class=ValueError,
        )

    @staticmethod
    def encode(timestamp_protobuf_object, timestamp_object: "Timestamp") -> None:
        """
        Encode an instance of this class into the protocol buffer object.

        The protocol buffer object in the timestamp_protobuf_object argument is matched with the instance of this class in the 'timestamp_object' argument.

        :param timestamp_protobuf_object: the protocol buffer object whose type corresponds with this class.
        :param timestamp_object: an instance of this class to be encoded in the protocol buffer object.
        """
        timestamp_protobuf_object.seconds = timestamp_object.seconds
        timestamp_protobuf_object.nanos = timestamp_object.nanos

    @classmethod
    def decode(cls, timestamp_protobuf_object) -> "Timestamp":
        """
        Decode a protocol buffer object that corresponds with this class into an instance of this class.

        A new instance of this class is created that matches the protocol buffer object in the 'timestamp_protobuf_object' argument.

        :param timestamp_protobuf_object: the protocol buffer object whose type corresponds with this class.
        :return: A new instance of this class that matches the protocol buffer object in the 'timestamp_protobuf_object' argument.
        """
        return Timestamp(
            timestamp_protobuf_object.seconds, timestamp_protobuf_object.nanos
        )

    def __eq__(self, other) -> bool:
        """Compare with another object."""
        return (
            isinstance(other, Timestamp)
            and self.seconds == other.seconds
            and self.nanos == other.nanos
        )


class PublicKey:
    """This class represents an instance of PublicKey."""

    class PublicKeyType(Enum):
        """Enumeration of public key types supported by Tendermint."""

        ed25519 = "ed25519"
        secp256k1 = "secp256k1"

    def __init__(self, data: bytes, key_type: PublicKeyType) -> None:
        """
        Initialize the public key object.

        :param data: the data of the public key.
        :param key_type: the type of the public key.
        """
        self.data = data
        self.key_type = key_type

    @staticmethod
    def encode(public_key_protobuf_object, public_key_object: "PublicKey") -> None:
        """
        Encode an instance of this class into the protocol buffer object.

        The protocol buffer object in the public_key_protobuf_object argument is matched with the instance of this class in the 'public_key_object' argument.

        :param public_key_protobuf_object: the protocol buffer object whose type corresponds with this class.
        :param public_key_object: an instance of this class to be encoded in the protocol buffer object.
        """
        key_type_name = public_key_object.key_type.value
        setattr(public_key_protobuf_object, key_type_name, public_key_object.data)

    @classmethod
    def decode(cls, public_key_protobuf_object) -> "PublicKey":
        """
        Decode a protocol buffer object that corresponds with this class into an instance of this class.

        A new instance of this class is created that matches the protocol buffer object in the 'public_key_protobuf_object' argument.

        :param public_key_protobuf_object: the protocol buffer object whose type corresponds with this class.
        :return: A new instance of this class that matches the protocol buffer object in the 'public_key_protobuf_object' argument.
        """
        key_type_name = public_key_protobuf_object.WhichOneof("sum")
        key_type = PublicKey.PublicKeyType(key_type_name)
        data = getattr(public_key_protobuf_object, key_type_name)
        return PublicKey(data, key_type)

    def __eq__(self, other):
        """Compare with another object."""
        return (
            isinstance(other, PublicKey)
            and self.data == other.data
            and self.key_type == other.key_type
        )


class ValidatorUpdate:
    """This class represents an instance of ValidatorUpdate."""

    def __init__(self, pub_key: PublicKey, power: int):
        """Initialise an instance of ValidatorUpdate."""
        self.pub_key = pub_key
        self.power = power

    @staticmethod
    def encode(
        validator_update_protobuf_object, validator_update_object: "ValidatorUpdate"
    ) -> None:
        """
        Encode an instance of this class into the protocol buffer object.

        The protocol buffer object in the validator_update_protobuf_object argument is matched with the instance of this class in the 'validator_update_object' argument.

        :param validator_update_protobuf_object: the protocol buffer object whose type corresponds with this class.
        :param validator_update_object: an instance of this class to be encoded in the protocol buffer object.
        """
        PublicKey.encode(
            validator_update_protobuf_object.pub_key, validator_update_object.pub_key
        )
        validator_update_protobuf_object.power = validator_update_object.power

    @classmethod
    def decode(cls, validator_update_protobuf_object) -> "ValidatorUpdate":
        """
        Decode a protocol buffer object that corresponds with this class into an instance of this class.

        A new instance of this class is created that matches the protocol buffer object in the 'validator_update_protobuf_object' argument.

        :param validator_update_protobuf_object: the protocol buffer object whose type corresponds with this class.
        :return: A new instance of this class that matches the protocol buffer object in the 'validator_update_protobuf_object' argument.
        """
        pub_key = PublicKey.decode(validator_update_protobuf_object.pub_key)
        return ValidatorUpdate(
            pub_key,
            validator_update_protobuf_object.power,
        )

    def __eq__(self, other):
        """Compare with another object."""
        return (
            isinstance(other, ValidatorUpdate)
            and self.pub_key == other.pub_key
            and self.power == other.power
        )


class ValidatorUpdates:
    """This class represents an instance of ValidatorUpdates."""

    def __init__(self, validator_updates: List[ValidatorUpdate]):
        """Initialise an instance of ValidatorUpdates."""
        self.validator_updates = validator_updates

    @staticmethod
    def encode(
        validator_updates_protobuf_object, validator_updates_object: "ValidatorUpdates"
    ) -> None:
        """
        Encode an instance of this class into the protocol buffer object.

        The protocol buffer object in the validator_updates_protobuf_object argument is matched with the instance of this class in the 'validator_updates_object' argument.

        :param validator_updates_protobuf_object: the protocol buffer object whose type corresponds with this class.
        :param validator_updates_object: an instance of this class to be encoded in the protocol buffer object.
        """
        validator_updates = []
        for validator_update_object in validator_updates_object.validator_updates:
            validator_update_protobuf_object = (
                abci_pb2.AbciMessage.ValidatorUpdates.ValidatorUpdate()
            )
            ValidatorUpdate.encode(
                validator_update_protobuf_object, validator_update_object
            )
            validator_updates.append(validator_update_protobuf_object)
        validator_updates_protobuf_object.validators.extend(validator_updates)

    @classmethod
    def decode(cls, validator_updates_protobuf_object) -> "ValidatorUpdates":
        """
        Decode a protocol buffer object that corresponds with this class into an instance of this class.

        A new instance of this class is created that matches the protocol buffer object in the 'validator_updates_protobuf_object' argument.

        :param validator_updates_protobuf_object: the protocol buffer object whose type corresponds with this class.
        :return: A new instance of this class that matches the protocol buffer object in the 'validator_updates_protobuf_object' argument.
        """
        validator_updates_objects = []
        validator_updates_protobuf_objects = list(
            validator_updates_protobuf_object.validators
        )
        for validator_update_protobuf_object in validator_updates_protobuf_objects:
            validator_update = ValidatorUpdate.decode(validator_update_protobuf_object)
            validator_updates_objects.append(validator_update)
        return ValidatorUpdates(validator_updates_objects)

    def __eq__(self, other):
        """Compare with another object."""
        return (
            isinstance(other, ValidatorUpdates)
            and self.validator_updates == other.validator_updates
        )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test the behaviours.py module of the skill."""
from enum import Enum
from typing import Any, Dict, Generator, Optional, Tuple
from unittest import mock
from unittest.mock import MagicMock

import pytest

from packages.valory.skills.abstract_round_abci.base import (
    ABCIAppInternalError,
    AbciApp,
    AbstractRound,
    BaseSynchronizedData,
    BaseTxPayload,
    DegenerateRound,
    EventType,
    RoundSequence,
)
from packages.valory.skills.abstract_round_abci.behaviour_utils import (
    BaseBehaviour,
    DegenerateBehaviour,
)
from packages.valory.skills.abstract_round_abci.behaviours import (
    AbstractRoundBehaviour,
    _MetaRoundBehaviour,
)


BEHAVIOUR_A_ID = "behaviour_a"
BEHAVIOUR_B_ID = "behaviour_b"
BEHAVIOUR_C_ID = "behaviour_c"
ROUND_A_ID = "round_a"
ROUND_B_ID = "round_b"


class RoundA(AbstractRound):
    """Round A."""

    round_id = ROUND_A_ID
    allowed_tx_type = "payload_a"

    def end_block(self) -> Optional[Tuple[BaseSynchronizedData, EventType]]:
        """End block."""

    def check_payload(self, payload: BaseTxPayload) -> None:
        """Check payload."""

    def process_payload(self, payload: BaseTxPayload) -> None:
        """Process payload."""


class RoundB(AbstractRound):
    """Round B."""

    round_id = ROUND_B_ID
    allowed_tx_type = "payload_b"

    def end_block(self) -> Optional[Tuple[BaseSynchronizedData, EventType]]:
        """End block."""

    def check_payload(self, payload: BaseTxPayload) -> None:
        """Check payload."""

    def process_payload(self, payload: BaseTxPayload) -> None:
        """Process payload."""


class BehaviourA(BaseBehaviour):
    """Dummy behaviour."""

    behaviour_id = BEHAVIOUR_A_ID
    matching_round = RoundA

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize behaviour."""
        super().__init__(*args, **kwargs)
        self.count = 0

    def setup(self) -> None:
        """Setup behaviour."""
        self.count += 1

    def async_act(self) -> Generator:
        """Dummy act method."""
        yield


class BehaviourB(BaseBehaviour):
    """Dummy behaviour."""

    behaviour_id = BEHAVIOUR_B_ID
    matching_round = RoundB

    def async_act(self) -> Generator:
        """Dummy act method."""
        yield


class ConcreteAbciApp(AbciApp):
    """Concrete ABCI App."""

    initial_round_cls = RoundA
    transition_function = {RoundA: {MagicMock(): RoundB}}
    event_to_timeout: Dict = {}


class ConcreteRoundBehaviour(AbstractRoundBehaviour):
    """Concrete round behaviour."""

    abci_app_cls = ConcreteAbciApp
    behaviours = {BehaviourA, BehaviourB}  # type: ignore
    initial_behaviour_cls = BehaviourA


class TestAbstractRoundBehaviour:
    """Test 'AbstractRoundBehaviour' class."""

    def setup(self) -> None:
        """Set up the tests."""
        self.round_sequence_mock = MagicMock()
        context_mock = MagicMock()
        context_mock.state.round_sequence = self.round_sequence_mock
        context_mock.state.round_sequence.syncing_up = False
        context_mock.params.ipfs_domain_name = None
        self.behaviour = ConcreteRoundBehaviour(name="", skill_context=context_mock)

    def test_setup(self) -> None:
        """Test 'setup' method."""
        self.behaviour.setup()

    def test_teardown(self) -> None:
        """Test 'teardown' method."""
        self.behaviour.teardown()

    def test_current_behaviour_return_none(self) -> None:
        """Test 'current_behaviour' property return None."""
        assert self.behaviour.current_behaviour is None

    def test_act_current_behaviour_name_is_none(self) -> None:
        """Test 'act' with current behaviour None."""
        self.behaviour.current_behaviour = None
        with mock.patch.object(self.behaviour, "_process_current_round"):
            self.behaviour.act()

    def test_check_matching_round_consistency(self) -> None:
        """Test classmethod '_get_behaviour_id_to_behaviour_mapping', negative case."""
        rounds = [MagicMock(round_id=f"round_{i}") for i in range(3)]
        mock_behaviours = [
            MagicMock(matching_round=round, behaviour_id=f"behaviour_{i}")
            for i, round in enumerate(rounds)
        ]

        with mock.patch(
            "packages.valory.skills.abstract_round_abci.behaviours._MetaRoundBehaviour._check_all_required_classattributes_are_set"
        ), mock.patch(
            "packages.valory.skills.abstract_round_abci.behaviours._MetaRoundBehaviour._check_behaviour_id_uniqueness"
        ), mock.patch(
            "packages.valory.skills.abstract_round_abci.behaviours._MetaRoundBehaviour._check_initial_behaviour_in_set_of_behaviours"
        ), pytest.raises(
            ABCIAppInternalError,
            match="internal error: round round_0 is a final round it shouldn't have any matching behaviours",
        ):

            class MyRoundBehaviour(AbstractRoundBehaviour):
                abci_app_cls = MagicMock(
                    get_all_round_classes=lambda: rounds,
                    final_states={
                        rounds[0],
                    },
                )
                behaviours = mock_behaviours  # type: ignore
                initial_behaviour_cls = MagicMock()

            MyRoundBehaviour(name=MagicMock(), skill_context=MagicMock())

    def test_get_behaviour_id_to_behaviour_mapping_negative(self) -> None:
        """Test classmethod '_get_behaviour_id_to_behaviour_mapping', negative case."""
        behaviour_id = "behaviour_id"
        behaviour_1 = MagicMock(behaviour_id=behaviour_id)
        behaviour_2 = MagicMock(behaviour_id=behaviour_id)

        with pytest.raises(
            ValueError,
            match=f"cannot have two behaviours with the same id; got {behaviour_2} and {behaviour_1} both with id '{behaviour_id}'",
        ):
            with mock.patch(
                "packages.valory.skills.abstract_round_abci.behaviours._MetaRoundBehaviour._check_consistency"
            ):

                class MyRoundBehaviour(AbstractRoundBehaviour):
                    abci_app_cls = MagicMock
                    behaviours = [behaviour_1, behaviour_2]  # type: ignore
                    initial_behaviour_cls = MagicMock()

                MyRoundBehaviour(name=MagicMock(), skill_context=MagicMock())

    def test_get_round_to_behaviour_mapping_two_behaviours_same_round(self) -> None:
        """Test classmethod '_get_round_to_behaviour_mapping' when two different behaviours point to the same round."""
        behaviour_id_1 = "behaviour_id_1"
        behaviour_id_2 = "behaviour_id_2"
        round_cls = RoundA
        round_id = round_cls.round_id
        behaviour_1 = MagicMock(behaviour_id=behaviour_id_1, matching_round=round_cls)
        behaviour_2 = MagicMock(behaviour_id=behaviour_id_2, matching_round=round_cls)

        with pytest.raises(
            ValueError,
            match=f"the behaviours '{behaviour_id_2}' and '{behaviour_id_1}' point to the same matching round '{round_id}'",
        ):
            with mock.patch(
                "packages.valory.skills.abstract_round_abci.behaviours._MetaRoundBehaviour._check_consistency"
            ):

                class MyRoundBehaviour(AbstractRoundBehaviour):
                    abci_app_cls = ConcreteAbciApp
                    behaviours = [behaviour_1, behaviour_2]  # type: ignore
                    initial_behaviour_cls = behaviour_1

                MyRoundBehaviour(name=MagicMock(), skill_context=MagicMock())

    def test_get_round_to_behaviour_mapping_with_final_rounds(self) -> None:
        """Test classmethod '_get_round_to_behaviour_mapping' with final rounds."""

        class FinalRound(DegenerateRound):
            """A final round for testing."""

            def check_payload(self, payload: BaseTxPayload) -> None:
                pass

            def process_payload(self, payload: BaseTxPayload) -> None:
                pass

            def end_block(self) -> Optional[Tuple[BaseSynchronizedData, Enum]]:
                pass

        behaviour_id_1 = "behaviour_id_1"
        behaviour_1 = MagicMock(behaviour_id=behaviour_id_1, matching_round=RoundA)

        class AbciAppTest(AbciApp):
            """Abci App for testing."""

            initial_round_cls = RoundA
            transition_function = {RoundA: {MagicMock(): FinalRound}, FinalRound: {}}
            event_to_timeout: Dict = {}
            final_states = {FinalRound}

        class MyRoundBehaviour(AbstractRoundBehaviour):
            abci_app_cls = AbciAppTest
            behaviours = [behaviour_1]  # type: ignore
            initial_behaviour_cls = behaviour_1

        behaviour = MyRoundBehaviour(name=MagicMock(), skill_context=MagicMock())
        final_behaviour = behaviour._round_to_behaviour[FinalRound]
        assert issubclass(final_behaviour, DegenerateBehaviour)
        assert final_behaviour.behaviour_id == f"degenerate_{FinalRound.round_id}"

    def test_check_behaviour_id_uniqueness_negative(self) -> None:
        """Test metaclass method '_check_consistency', negative case."""
        behaviour_id = "behaviour_id"
        behaviour_1_cls_name = "Behaviour1"
        behaviour_2_cls_name = "Behaviour2"
        behaviour_1 = MagicMock(
            behaviour_id=behaviour_id, __name__=behaviour_1_cls_name
        )
        behaviour_2 = MagicMock(
            behaviour_id=behaviour_id, __name__=behaviour_2_cls_name
        )

        with pytest.raises(
            ABCIAppInternalError,
            match=fr"behaviours \['{behaviour_1_cls_name}', '{behaviour_2_cls_name}'\] have the same behaviour id '{behaviour_id}'",
        ):

            class MyRoundBehaviour(AbstractRoundBehaviour):
                abci_app_cls = MagicMock
                behaviours = [behaviour_1, behaviour_2]  # type: ignore
                initial_behaviour_cls = MagicMock()

    def test_check_consistency_two_behaviours_same_round(self) -> None:
        """Test metaclass method '_check_consistency' when two different behaviours point to the same round."""
        behaviour_id_1 = "behaviour_id_1"
        behaviour_id_2 = "behaviour_id_2"
        round_cls = RoundA
        round_id = round_cls.round_id
        behaviour_1 = MagicMock(behaviour_id=behaviour_id_1, matching_round=round_cls)
        behaviour_2 = MagicMock(behaviour_id=behaviour_id_2, matching_round=round_cls)

        with pytest.raises(
            ABCIAppInternalError,
            match=rf"internal error: behaviours \['{behaviour_id_1}', '{behaviour_id_2}'\] have the same matching round '{round_id}'",
        ):

            class MyRoundBehaviour(AbstractRoundBehaviour):
                abci_app_cls = ConcreteAbciApp
                behaviours = [behaviour_1, behaviour_2]  # type: ignore
                initial_behaviour_cls = behaviour_1

    def test_check_initial_behaviour_in_set_of_behaviours_negative_case(self) -> None:
        """Test classmethod '_check_initial_behaviour_in_set_of_behaviours' when initial behaviour is NOT in the set."""
        behaviour_1 = MagicMock(
            behaviour_id="behaviour_id_1", matching_round=MagicMock()
        )
        behaviour_2 = MagicMock(
            behaviour_id="behaviour_id_2", matching_round=MagicMock()
        )

        with pytest.raises(
            ABCIAppInternalError,
            match="initial behaviour behaviour_id_2 is not in the set of behaviours",
        ):

            class MyRoundBehaviour(AbstractRoundBehaviour):
                abci_app_cls = ConcreteAbciApp
                behaviours = [behaviour_1]  # type: ignore
                initial_behaviour_cls = behaviour_2

    def test_act_no_round_change(self) -> None:
        """Test the 'act' method of the behaviour, with no round change."""
        self.round_sequence_mock.current_round = RoundA(MagicMock(), MagicMock())
        self.round_sequence_mock.current_round_height = 0

        # check that after setup(), current behaviour is initial behaviour
        self.behaviour.setup()
        assert isinstance(self.behaviour.current_behaviour, BehaviourA)

        # check that after act(), current behaviour is initial behaviour
        self.behaviour.act()
        assert isinstance(self.behaviour.current_behaviour, BehaviourA)

        # check that once the flag done is set, tries to schedule
        # the next behaviour, but without success
        self.behaviour.current_behaviour.set_done()
        self.behaviour.act()
        assert self.behaviour.current_behaviour is None

    def test_act_behaviour_setup(self) -> None:
        """Test the 'act' method of the FSM behaviour triggers setup() of the behaviour."""
        self.round_sequence_mock.current_round = RoundA(MagicMock(), MagicMock())
        self.round_sequence_mock.current_round_height = 0

        # check that after setup(), current behaviour is initial behaviour
        self.behaviour.setup()
        assert isinstance(self.behaviour.current_behaviour, BehaviourA)

        assert self.behaviour.current_behaviour.count == 0

        # check that after act() first time, a call to setup has been made
        self.behaviour.act()
        assert isinstance(self.behaviour.current_behaviour, BehaviourA)
        assert self.behaviour.current_behaviour.count == 1

        # check that after act() second time, no further call to setup
        self.behaviour.act()
        assert self.behaviour.current_behaviour.count == 1

    def test_act_with_round_change(self) -> None:
        """Test the 'act' method of the behaviour, with round change."""
        self.round_sequence_mock.current_round = RoundA(MagicMock(), MagicMock())
        self.round_sequence_mock.current_round_height = 0

        # check that after setup(), current behaviour is initial behaviour
        self.behaviour.setup()
        assert isinstance(self.behaviour.current_behaviour, BehaviourA)

        # check that after act(), current behaviour is initial behaviour
        self.behaviour.act()
        assert isinstance(self.behaviour.current_behaviour, BehaviourA)

        # change the round
        self.round_sequence_mock.current_round = RoundB(MagicMock(), MagicMock())
        self.round_sequence_mock.current_round_height = (
            self.round_sequence_mock.current_round_height + 1
        )

        # check that if the round is changed, the behaviour transition is taken
        self.behaviour.act()
        assert isinstance(self.behaviour.current_behaviour, BehaviourB)

    def test_act_with_round_change_after_current_behaviour_is_none(self) -> None:
        """Test the 'act' method of the behaviour, with round change, after cur behaviour is none."""
        self.round_sequence_mock.current_round = RoundA(MagicMock(), MagicMock())
        self.round_sequence_mock.current_round_height = 0

        # instantiate behaviour
        self.behaviour.current_behaviour = self.behaviour.instantiate_behaviour_cls(BehaviourA)  # type: ignore

        # check that after act(), current behaviour is same behaviour
        self.behaviour.act()
        assert isinstance(self.behaviour.current_behaviour, BehaviourA)

        # check that after the behaviour is done, current behaviour is None
        self.behaviour.current_behaviour.set_done()
        self.behaviour.act()
        assert self.behaviour.current_behaviour is None

        # change the round
        self.round_sequence_mock.current_round = RoundB(MagicMock(), MagicMock())
        self.round_sequence_mock.current_round_height = (
            self.round_sequence_mock.current_round_height + 1
        )

        # check that if the round is changed, the behaviour transition is taken
        self.behaviour.act()
        assert isinstance(self.behaviour.current_behaviour, BehaviourB)


def test_meta_round_behaviour_when_instance_not_subclass_of_abstract_round() -> None:
    """Test instantiation of meta class when instance not a subclass of abstract round."""

    class MyRoundBeahviour(metaclass=_MetaRoundBehaviour):
        pass


def test_abstract_round_behaviour_instantiation_without_attributes_raises_error() -> None:
    """Test that definition of concrete subclass of AbstractRoundBehavior without attributes raises error."""
    with pytest.raises(ABCIAppInternalError):

        class MyRoundBehaviour(AbstractRoundBehaviour):
            pass


def test_abstract_round_behaviour_matching_rounds_not_covered() -> None:
    """Test that definition of concrete subclass of AbstractRoundBehavior when matching round not covered."""
    with pytest.raises(ABCIAppInternalError):

        class MyRoundBehaviour(AbstractRoundBehaviour):
            abci_app_cls = ConcreteAbciApp
            behaviours = {BehaviourA}  # type: ignore
            initial_behaviour_cls = BehaviourA


def test_self_loops_in_abci_app_reinstantiate_behaviour() -> None:
    """Test that a self-loop transition in the AbciApp will trigger a transition in the round behaviour."""
    event = MagicMock()

    class AbciAppTest(AbciApp):
        initial_round_cls = RoundA
        transition_function = {RoundA: {event: RoundA}}

    class RoundBehaviour(AbstractRoundBehaviour):
        abci_app_cls = AbciAppTest
        behaviours = {BehaviourA}  # type: ignore
        initial_behaviour_cls = BehaviourA

    round_sequence = RoundSequence(AbciAppTest)
    round_sequence.end_sync()
    round_sequence.setup(MagicMock(), MagicMock(), MagicMock())
    context_mock = MagicMock()
    context_mock.state.round_sequence = round_sequence
    context_mock.params.ipfs_domain_name = None
    behaviour = RoundBehaviour(name="", skill_context=context_mock)
    behaviour.setup()

    behaviour_1 = behaviour.current_behaviour
    assert isinstance(behaviour_1, BehaviourA)

    round_sequence.abci_app.process_event(event)

    behaviour.act()
    behaviour_2 = behaviour.current_behaviour
    assert isinstance(behaviour_2, BehaviourA)
    assert id(behaviour_1) != id(behaviour_2)
    assert behaviour_1 != behaviour_2



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the shared state for the price estimation ABCI application."""
import inspect
import json
from pathlib import Path
from time import time
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, cast

from aea.exceptions import enforce
from aea.skills.base import Model

from packages.valory.protocols.http.message import HttpMessage
from packages.valory.skills.abstract_round_abci.base import (
    AbciApp,
    AbciAppDB,
    BaseSynchronizedData,
    ConsensusParams,
    RoundSequence,
)


NUMBER_OF_RETRIES: int = 5
_DEFAULT_REQUEST_TIMEOUT = 10.0
_DEFAULT_REQUEST_RETRY_DELAY = 1.0
_DEFAULT_TX_TIMEOUT = 10.0
_DEFAULT_TX_MAX_ATTEMPTS = 10


class BaseParams(Model):  # pylint: disable=too-many-instance-attributes
    """Parameters."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize the parameters object."""
        self.service_id = self._ensure("service_id", kwargs)
        self.tendermint_url = self._ensure("tendermint_url", kwargs)
        self.max_healthcheck = self._ensure("max_healthcheck", kwargs)
        self.round_timeout_seconds = self._ensure("round_timeout_seconds", kwargs)
        self.sleep_time = self._ensure("sleep_time", kwargs)
        self.retry_timeout = self._ensure("retry_timeout", kwargs)
        self.retry_attempts = self._ensure("retry_attempts", kwargs)
        self.keeper_timeout = self._ensure("keeper_timeout", kwargs)
        self.observation_interval = self._ensure("observation_interval", kwargs)
        self.drand_public_key = self._ensure("drand_public_key", kwargs)
        self.tendermint_com_url = self._ensure("tendermint_com_url", kwargs)
        self.tendermint_max_retries = self._ensure("tendermint_max_retries", kwargs)
        self.tendermint_check_sleep_delay = self._ensure(
            "tendermint_check_sleep_delay", kwargs
        )
        self.reset_tendermint_after = self._ensure("reset_tendermint_after", kwargs)
        self.consensus_params = ConsensusParams.from_json(kwargs.pop("consensus", {}))
        self.cleanup_history_depth = self._ensure("cleanup_history_depth", kwargs)
        self.request_timeout = kwargs.pop("request_timeout", _DEFAULT_REQUEST_TIMEOUT)
        self.request_retry_delay = kwargs.pop(
            "request_retry_delay", _DEFAULT_REQUEST_RETRY_DELAY
        )
        self.tx_timeout = kwargs.pop("tx_timeout", _DEFAULT_TX_TIMEOUT)
        self.max_attempts = kwargs.pop("max_attempts", _DEFAULT_TX_MAX_ATTEMPTS)
        setup_params = kwargs.pop("setup", {})
        # we sanitize for null values as these are just kept for schema definitions
        setup_params = {
            key: val for key, val in setup_params.items() if val is not None
        }
        self.setup_params = setup_params
        super().__init__(*args, **kwargs)

    @classmethod
    def _ensure(cls, key: str, kwargs: Dict) -> Any:
        """Get and ensure the configuration field is not None."""
        value = kwargs.pop(key, None)
        enforce(value is not None, f"'{key}' required, but it is not set")
        return value


class SharedState(Model):
    """Keep the current shared state of the skill."""

    def __init__(self, *args: Any, abci_app_cls: Type[AbciApp], **kwargs: Any) -> None:
        """Initialize the state."""
        self.abci_app_cls = self._process_abci_app_cls(abci_app_cls)
        self._round_sequence: Optional[RoundSequence] = None
        super().__init__(*args, **kwargs)

    def setup(self) -> None:
        """Set up the model."""
        self._round_sequence = RoundSequence(self.abci_app_cls)
        consensus_params = cast(BaseParams, self.context.params).consensus_params
        setup_params = cast(BaseParams, self.context.params).setup_params
        self.round_sequence.setup(
            BaseSynchronizedData(
                AbciAppDB(
                    setup_data=setup_params,
                    cross_period_persisted_keys=self.abci_app_cls.cross_period_persisted_keys,
                )
            ),
            consensus_params,
            self.context.logger,
        )

    @property
    def round_sequence(self) -> RoundSequence:
        """Get the round_sequence."""
        if self._round_sequence is None:
            raise ValueError("round sequence not available")
        return self._round_sequence

    @property
    def synchronized_data(self) -> BaseSynchronizedData:
        """Get the latest synchronized_data if available."""
        return self.round_sequence.latest_synchronized_data

    @classmethod
    def _process_abci_app_cls(cls, abci_app_cls: Type[AbciApp]) -> Type[AbciApp]:
        """Process the 'initial_round_cls' parameter."""
        if not inspect.isclass(abci_app_cls):
            raise ValueError(f"The object {abci_app_cls} is not a class")
        if not issubclass(abci_app_cls, AbciApp):
            cls_name = AbciApp.__name__
            cls_module = AbciApp.__module__
            raise ValueError(
                f"The class {abci_app_cls} is not an instance of {cls_module}.{cls_name}"
            )
        return abci_app_cls


class Requests(Model):
    """Keep the current pending requests."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize the state."""
        super().__init__(*args, **kwargs)

        # mapping from dialogue reference nonce to a callback
        self.request_id_to_callback: Dict[str, Callable] = {}


class ApiSpecs(Model):  # pylint: disable=too-many-instance-attributes
    """A model that wraps APIs to get cryptocurrency prices."""

    url: str
    api_id: str
    method: str
    response_key: str
    response_type: str
    headers: List[Tuple[str, str]]
    parameters: List[Tuple[str, str]]

    _retries_attempted: int
    _retries: int
    _response_types: Dict[str, Any] = {
        "int": int,
        "float": float,
        "dict": dict,
        "list": list,
        "str": str,
    }

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize ApiSpecsModel."""

        self.url = self.ensure("url", kwargs)
        self.api_id = self.ensure("api_id", kwargs)
        self.method = self.ensure("method", kwargs)
        self.headers = kwargs.pop("headers", [])
        self.parameters = kwargs.pop("parameters", [])
        self.response_key = kwargs.pop("response_key", None)
        self.response_type = kwargs.pop("response_type", str)

        self._retries_attempted = 0
        self._retries = kwargs.pop("retries", NUMBER_OF_RETRIES)

        super().__init__(*args, **kwargs)

    def ensure(self, keyword: str, kwargs: Dict) -> Any:
        """Ensure a keyword argument."""
        value = kwargs.pop(keyword, None)
        if value is None:
            raise ValueError(
                f"Value for {keyword} is required by {self.__class__.__name__}."
            )
        return value

    def get_spec(
        self,
    ) -> Dict:
        """Returns dictionary containing api specifications."""

        return {
            "url": self.url,
            "method": self.method,
            "headers": self.headers,
            "parameters": self.parameters,
        }

    def process_response(self, response: HttpMessage) -> Any:
        """Process response from api."""
        try:
            response_data = json.loads(response.body.decode())
            if self.response_key is None:
                return response_data

            first_key, *keys = self.response_key.split(":")
            value = response_data[first_key]
            for key in keys:
                value = value[key]

            return self._response_types.get(self.response_type)(value)  # type: ignore

        except (json.JSONDecodeError, KeyError):
            return None

    def increment_retries(self) -> None:
        """Increment the retries counter."""
        self._retries_attempted += 1

    def reset_retries(self) -> None:
        """Reset the retries counter."""
        self._retries_attempted = 0

    def is_retries_exceeded(self) -> bool:
        """Check if the retries amount has been exceeded."""
        return self._retries_attempted > self._retries


class BenchmarkBlockTypes:  # pylint: disable=too-few-public-methods
    """Benchmark block types."""

    LOCAL = "local"
    CONSENSUS = "consensus"
    TOTAL = "total"


class BenchmarkBlock:
    """
    Benchmark

    This class represents logic to measure the code block using a
    context manager.
    """

    start: float
    total_time: float
    block_type: str

    def __init__(self, block_type: str) -> None:
        """Benchmark for single round."""
        self.block_type = block_type
        self.start = 0
        self.total_time = 0

    def __enter__(
        self,
    ) -> None:
        """Enter context."""
        self.start = time()

    def __exit__(self, *args: List, **kwargs: Dict) -> None:
        """Exit context"""
        self.total_time = time() - self.start


class BenchmarkBehaviour:
    """
    BenchmarkBehaviour

    This class represents logic to benchmark a single behaviour.
    """

    local_data: Dict[str, BenchmarkBlock]

    def __init__(
        self,
    ) -> None:
        """Initialize Benchmark behaviour object."""
        self.local_data = {}

    def _measure(self, block_type: str) -> BenchmarkBlock:
        """
        Returns a BenchmarkBlock object.

        :param block_type: type of block (e.g. local, consensus, request)
        :return: BenchmarkBlock
        """

        if block_type not in self.local_data:
            self.local_data[block_type] = BenchmarkBlock(block_type)

        return self.local_data[block_type]

    def local(
        self,
    ) -> BenchmarkBlock:
        """Measure local block."""
        return self._measure(BenchmarkBlockTypes.LOCAL)

    def consensus(
        self,
    ) -> BenchmarkBlock:
        """Measure consensus block."""
        return self._measure(BenchmarkBlockTypes.CONSENSUS)


class BenchmarkTool(Model):
    """
    BenchmarkTool

    Tool to benchmark ABCI apps.
    """

    benchmark_data: Dict[str, BenchmarkBehaviour]
    log_dir: Path

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Benchmark tool for rounds behaviours."""
        super().__init__(*args, **kwargs)
        self.benchmark_data = {}
        self.log_dir = Path(kwargs.pop("log_dir", "/logs"))

    def measure(self, behaviour: str) -> BenchmarkBehaviour:
        """Measure time to complete round."""
        if behaviour not in self.benchmark_data:
            self.benchmark_data[behaviour] = BenchmarkBehaviour()
        return self.benchmark_data[behaviour]

    @property
    def data(
        self,
    ) -> List:
        """Returns formatted data."""

        behavioural_data = []
        for behaviour, tool in self.benchmark_data.items():
            data = {k: v.total_time for k, v in tool.local_data.items()}
            data[BenchmarkBlockTypes.TOTAL] = sum(data.values())
            behavioural_data.append({"behaviour": behaviour, "data": data})

        return behavioural_data

    def save(self, period: int = 0, reset: bool = True) -> None:
        """Save logs to a file."""

        try:
            self.log_dir.mkdir(exist_ok=True)
            agent_dir = self.log_dir / self.context.agent_address
            agent_dir.mkdir(exist_ok=True)
            filepath = agent_dir / f"{period}.json"

            with open(str(filepath), "w+", encoding="utf-8") as outfile:
                json.dump(self.data, outfile)
            self.context.logger.info(f"Saving benchmarking data for period: {period}")

        except PermissionError as e:  # pragma: nocover
            self.context.logger.info(f"Error saving benchmark data:\n{e}")

        if reset:
            self.reset()

    def reset(
        self,
    ) -> None:
        """Reset benchmark data"""
        self.benchmark_data = {}



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test forecasting operations."""
import re
from copy import deepcopy
from typing import Any

import numpy as np
import pandas as pd
import pytest
from _pytest.monkeypatch import MonkeyPatch
from pmdarima import ARIMA
from pmdarima.pipeline import Pipeline
from pmdarima.preprocessing import FourierFeaturizer

from packages.valory.skills.apy_estimation_abci.ml.forecasting import (
    PoolIdToTrainDataType,
    baseline,
    calc_metrics,
    estimate_apy_per_pool,
    init_forecaster,
    predict_safely,
    report_metrics,
)
from packages.valory.skills.apy_estimation_abci.ml.forecasting import (
    test_forecaster as _test_forecaster,
)
from packages.valory.skills.apy_estimation_abci.ml.forecasting import (
    test_forecaster_per_pool as _test_forecaster_per_pool,
)
from packages.valory.skills.apy_estimation_abci.ml.forecasting import (
    train_forecaster,
    train_forecaster_per_pool,
    update_forecaster_per_pool,
    walk_forward_test,
)

from tests.test_skills.test_apy_estimation_abci.conftest import DummyPipeline


class TestForecasting:
    """Tests for forecasting operations."""

    def setup(self) -> None:
        """Initialize class."""
        self._pipeline = DummyPipeline()

    @staticmethod
    def test_init_forecaster(hyperparameters: Any) -> None:
        """Test `init_forecaster`."""
        forecaster = init_forecaster(**hyperparameters)
        pipeline_steps = forecaster.get_params()["steps"]
        fourier_actual = pipeline_steps[0][1].get_params()
        arima_actual = pipeline_steps[1][1].get_params()

        fourier_expected = {
            "k": hyperparameters["k"],
            "m": hyperparameters["m"],
            "prefix": None,
        }
        arima_expected = {
            "maxiter": 150,
            "method": "lbfgs",
            "order": (hyperparameters["p"], hyperparameters["q"], hyperparameters["d"]),
            "out_of_sample_size": 0,
            "scoring": "mse",
            "scoring_args": None,
            "seasonal_order": (0, 0, 0, 0),
            "start_params": None,
            "suppress_warnings": True,
            "trend": None,
            "with_intercept": True,
        }

        assert fourier_actual == fourier_expected
        assert arima_actual == arima_expected

    @staticmethod
    def test_train_forecaster(observations: np.ndarray, hyperparameters: Any) -> None:
        """Test `train_forecaster`."""
        forecaster = init_forecaster(**hyperparameters)
        forecaster_trained = train_forecaster(observations, **hyperparameters)
        assert type(forecaster) == type(
            forecaster_trained
        )  # must return the same type of object
        assert forecaster is not forecaster_trained  # object identity MUST be different

    @staticmethod
    def test_train_forecaster_per_pool(
        train_task_input: PoolIdToTrainDataType,
        trained_forecaster: Pipeline,
        monkeypatch: MonkeyPatch,
    ) -> None:
        """Test `train_forecaster_per_pool`."""
        monkeypatch.setattr(
            "packages.valory.skills.apy_estimation_abci.ml.forecasting.train_forecaster",
            lambda _, **__: trained_forecaster,
        )
        forecasters = train_forecaster_per_pool(
            train_task_input, dict.fromkeys(train_task_input, {})
        )
        assert len(forecasters) == len(train_task_input)
        for (id_actual, forecaster_actual), id_expected in zip(
            forecasters.items(), train_task_input.keys()
        ):
            assert id_actual == id_expected.replace(".csv", ".joblib")
            assert type(forecaster_actual) == Pipeline
            # must return the trained forecaster received from `train_forecaster`
            assert forecaster_actual == trained_forecaster

    @staticmethod
    def test_baseline() -> None:
        """Test `baseline`."""
        expected = np.asarray([0.1, 0.2, 0.3])
        actual = baseline(0.1, np.asarray([0.2, 0.3, 0.4]))
        assert np.array_equal(expected, actual)

    @staticmethod
    def test_calc_metrics(monkeypatch: MonkeyPatch) -> None:
        """Test `calc_metrics`."""
        metrics_to_patch = (
            "mean_pinball_loss",
            "smape",
            "explained_variance_score",
            "max_error",
            "mean_squared_error",
        )
        for metric_to_patch in metrics_to_patch:
            monkeypatch.setattr(
                f"packages.valory.skills.apy_estimation_abci.ml.forecasting.{metric_to_patch}",
                lambda *_: 0,
            )

        expected = {
            "mean pinball loss": 0,
            "SMAPE": 0,
            "Explained Variance": 0,
            "Max Error": 0,
            "MSE": 0,
        }

        actual = calc_metrics(np.empty(0), np.empty(0))

        assert actual == expected

    @staticmethod
    def test_report_metrics(monkeypatch: MonkeyPatch) -> None:
        """Test `report_metrics`."""
        metrics_res = {
            "mean pinball loss": 0,
            "SMAPE": 0,
            "Explained Variance": 0,
            "Max Error": 0,
            "MSE": 0,
        }
        monkeypatch.setattr(
            "packages.valory.skills.apy_estimation_abci.ml.forecasting.calc_metrics",
            lambda *_: metrics_res,
        )
        report = report_metrics(np.empty(0), np.empty(0), "test_name")
        assert isinstance(report, str)
        assert "test_name" in report
        assert any(metric_name in report for metric_name in metrics_res.keys())
        assert any(
            str(metric_value) in report for metric_value in set(metrics_res.values())
        )

    @pytest.mark.parametrize("steps_forward", (-2, 1, 6, 8))
    def test_walk_forward_test(self, steps_forward: int) -> None:
        """Test `walk_forward_test`."""
        test_size = 5
        params = self._pipeline, np.empty(test_size), steps_forward
        expected_output_len = test_size if steps_forward <= test_size else steps_forward
        if steps_forward < 0:
            with pytest.raises(
                ValueError, match="Timesteps to predict in the future cannot be -2 < 1."
            ):
                walk_forward_test(*params)
        else:
            if expected_output_len > test_size:
                with pytest.warns(
                    UserWarning,
                    match="Timesteps to predict in the future are larger than the number of test samples "
                    f"while using the Direct Multi-step Forecast Strategy: {steps_forward} > {test_size}",
                ):
                    predictions = walk_forward_test(*params)
            else:
                predictions = walk_forward_test(*params)

            assert isinstance(predictions, np.ndarray)
            assert len(predictions) == expected_output_len

    def test_test_forecaster_per_pool(
        self,
        train_task_input: PoolIdToTrainDataType,
        trained_forecaster: Pipeline,
        monkeypatch: MonkeyPatch,
    ) -> None:
        """Test `_test_forecaster_per_pool`."""
        monkeypatch.setattr(
            "packages.valory.skills.apy_estimation_abci.ml.forecasting.test_forecaster",
            lambda forecaster_, y_train_, y_test_, pair_name, _: {pair_name: "test"},
        )
        dummy_forecasters = {id_: trained_forecaster for id_ in train_task_input.keys()}
        _test_forecaster_per_pool(dummy_forecasters, train_task_input, train_task_input)

    def test_test_forecaster(self, monkeypatch: MonkeyPatch) -> None:
        """Test `test_forecaster`."""
        for testing_method in ("baseline", "walk_forward_test"):
            monkeypatch.setattr(
                f"packages.valory.skills.apy_estimation_abci.ml.forecasting.{testing_method}",
                lambda *_: None,
            )

        monkeypatch.setattr(
            "packages.valory.skills.apy_estimation_abci.ml.forecasting.report_metrics",
            lambda *_: "Report results.",
        )

        actual = _test_forecaster(self._pipeline, np.empty((5, 2)), np.empty(2), "test")

        expected = {}
        for testing_model in ("Baseline", "ARIMA"):
            expected[testing_model] = "Report results."

        assert actual == expected

    @pytest.mark.parametrize("id_mismatch", (True, False))
    def test_update_forecaster_per_pool(
        self,
        prepare_batch_task_result: pd.DataFrame,
        monkeypatch: MonkeyPatch,
        id_mismatch: bool,
    ) -> None:
        """Test `update_forecaster_per_pool`."""
        mismatch = "test" if id_mismatch else ""
        dummy_pipelines = {
            pool_id + mismatch: deepcopy(self._pipeline)
            for pool_id in prepare_batch_task_result["id"].values
        }

        if mismatch:
            update_forecaster_per_pool(prepare_batch_task_result, dummy_pipelines)
            assert not any(pipeline.updated for pipeline in dummy_pipelines.values())

        else:
            update_forecaster_per_pool(prepare_batch_task_result, dummy_pipelines)
            assert all(pipeline.updated for pipeline in dummy_pipelines.values())

    @pytest.mark.parametrize("steps_forward", (0, 1, 5))
    def test_estimate_apy_per_pool(
        self,
        monkeypatch: MonkeyPatch,
        steps_forward: int,
    ) -> None:
        """Test `estimate_apy_per_pool`."""
        dummy_pipelines = {
            f"pool{pool_id}": deepcopy(self._pipeline) for pool_id in range(3)
        }

        estimates = estimate_apy_per_pool(dummy_pipelines, steps_forward)
        expected_estimates = pd.DataFrame(
            {f"pool{pool_id}": np.ones(steps_forward) for pool_id in range(3)},
            index=[f"Step{i + 1} into the future" for i in range(steps_forward)],
        )
        pd.testing.assert_frame_equal(estimates, expected_estimates)

    def test_predict_safely(self) -> None:
        """
        Test `predict_safely`.

        Also prove that it is useful because of https://github.com/alkaline-ml/pmdarima/issues/404.
        """
        model = Pipeline(
            steps=[
                ("fourier", FourierFeaturizer(k=7, m=18)),
                (
                    "arima",
                    ARIMA(maxiter=150, order=(2, 3, 1), suppress_warnings=True),
                ),
            ]
        )
        y = np.array(
            [
                54.45259405,
                39.58028345,
                112.03066299,
                83.96699996,
                71.94854887,
                64.0529859,
                62.5824148,
                63.28134798,
                58.36879572,
                64.20712075,
                47.35538343,
                42.33252369,
                31.86261839,
                109.74997567,
                51.63416935,
                36.53481234,
                39.73377418,
                50.0544746,
                39.7898384,
                35.02107716,
                49.87511512,
            ]
        )
        steps_forward = 1

        # Fit model with data.
        model.fit(y)

        # Prove that the `pmdarima` would raise.
        with pytest.raises(
            ValueError,
            match=re.escape(
                "Input contains NaN, infinity or a value too large for dtype('float64')."
            ),
        ):
            model.predict(steps_forward)

        # Prove that `predict_safely` works as intended.
        y_hat = predict_safely(model, steps_forward)
        assert np.isnan(y_hat)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 valory
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains http's message definition."""

# pylint: disable=too-many-statements,too-many-locals,no-member,too-few-public-methods,too-many-branches,not-an-iterable,unidiomatic-typecheck,unsubscriptable-object
import logging
from typing import Any, Set, Tuple, cast

from aea.configurations.base import PublicId
from aea.exceptions import AEAEnforceError, enforce
from aea.protocols.base import Message


_default_logger = logging.getLogger("aea.packages.valory.protocols.http.message")

DEFAULT_BODY_SIZE = 4


class HttpMessage(Message):
    """A protocol for HTTP requests and responses."""

    protocol_id = PublicId.from_str("valory/http:1.0.0")
    protocol_specification_id = PublicId.from_str("valory/http:1.0.0")

    class Performative(Message.Performative):
        """Performatives for the http protocol."""

        REQUEST = "request"
        RESPONSE = "response"

        def __str__(self) -> str:
            """Get the string representation."""
            return str(self.value)

    _performatives = {"request", "response"}
    __slots__: Tuple[str, ...] = tuple()

    class _SlotsCls:
        __slots__ = (
            "body",
            "dialogue_reference",
            "headers",
            "message_id",
            "method",
            "performative",
            "status_code",
            "status_text",
            "target",
            "url",
            "version",
        )

    def __init__(
        self,
        performative: Performative,
        dialogue_reference: Tuple[str, str] = ("", ""),
        message_id: int = 1,
        target: int = 0,
        **kwargs: Any,
    ):
        """
        Initialise an instance of HttpMessage.

        :param message_id: the message id.
        :param dialogue_reference: the dialogue reference.
        :param target: the message target.
        :param performative: the message performative.
        :param **kwargs: extra options.
        """
        super().__init__(
            dialogue_reference=dialogue_reference,
            message_id=message_id,
            target=target,
            performative=HttpMessage.Performative(performative),
            **kwargs,
        )

    @property
    def valid_performatives(self) -> Set[str]:
        """Get valid performatives."""
        return self._performatives

    @property
    def dialogue_reference(self) -> Tuple[str, str]:
        """Get the dialogue_reference of the message."""
        enforce(self.is_set("dialogue_reference"), "dialogue_reference is not set.")
        return cast(Tuple[str, str], self.get("dialogue_reference"))

    @property
    def message_id(self) -> int:
        """Get the message_id of the message."""
        enforce(self.is_set("message_id"), "message_id is not set.")
        return cast(int, self.get("message_id"))

    @property
    def performative(self) -> Performative:  # type: ignore # noqa: F821
        """Get the performative of the message."""
        enforce(self.is_set("performative"), "performative is not set.")
        return cast(HttpMessage.Performative, self.get("performative"))

    @property
    def target(self) -> int:
        """Get the target of the message."""
        enforce(self.is_set("target"), "target is not set.")
        return cast(int, self.get("target"))

    @property
    def body(self) -> bytes:
        """Get the 'body' content from the message."""
        enforce(self.is_set("body"), "'body' content is not set.")
        return cast(bytes, self.get("body"))

    @property
    def headers(self) -> str:
        """Get the 'headers' content from the message."""
        enforce(self.is_set("headers"), "'headers' content is not set.")
        return cast(str, self.get("headers"))

    @property
    def method(self) -> str:
        """Get the 'method' content from the message."""
        enforce(self.is_set("method"), "'method' content is not set.")
        return cast(str, self.get("method"))

    @property
    def status_code(self) -> int:
        """Get the 'status_code' content from the message."""
        enforce(self.is_set("status_code"), "'status_code' content is not set.")
        return cast(int, self.get("status_code"))

    @property
    def status_text(self) -> str:
        """Get the 'status_text' content from the message."""
        enforce(self.is_set("status_text"), "'status_text' content is not set.")
        return cast(str, self.get("status_text"))

    @property
    def url(self) -> str:
        """Get the 'url' content from the message."""
        enforce(self.is_set("url"), "'url' content is not set.")
        return cast(str, self.get("url"))

    @property
    def version(self) -> str:
        """Get the 'version' content from the message."""
        enforce(self.is_set("version"), "'version' content is not set.")
        return cast(str, self.get("version"))

    def _is_consistent(self) -> bool:
        """Check that the message follows the http protocol."""
        try:
            enforce(
                isinstance(self.dialogue_reference, tuple),
                "Invalid type for 'dialogue_reference'. Expected 'tuple'. Found '{}'.".format(
                    type(self.dialogue_reference)
                ),
            )
            enforce(
                isinstance(self.dialogue_reference[0], str),
                "Invalid type for 'dialogue_reference[0]'. Expected 'str'. Found '{}'.".format(
                    type(self.dialogue_reference[0])
                ),
            )
            enforce(
                isinstance(self.dialogue_reference[1], str),
                "Invalid type for 'dialogue_reference[1]'. Expected 'str'. Found '{}'.".format(
                    type(self.dialogue_reference[1])
                ),
            )
            enforce(
                type(self.message_id) is int,
                "Invalid type for 'message_id'. Expected 'int'. Found '{}'.".format(
                    type(self.message_id)
                ),
            )
            enforce(
                type(self.target) is int,
                "Invalid type for 'target'. Expected 'int'. Found '{}'.".format(
                    type(self.target)
                ),
            )

            # Light Protocol Rule 2
            # Check correct performative
            enforce(
                isinstance(self.performative, HttpMessage.Performative),
                "Invalid 'performative'. Expected either of '{}'. Found '{}'.".format(
                    self.valid_performatives, self.performative
                ),
            )

            # Check correct contents
            actual_nb_of_contents = len(self._body) - DEFAULT_BODY_SIZE
            expected_nb_of_contents = 0
            if self.performative == HttpMessage.Performative.REQUEST:
                expected_nb_of_contents = 5
                enforce(
                    isinstance(self.method, str),
                    "Invalid type for content 'method'. Expected 'str'. Found '{}'.".format(
                        type(self.method)
                    ),
                )
                enforce(
                    isinstance(self.url, str),
                    "Invalid type for content 'url'. Expected 'str'. Found '{}'.".format(
                        type(self.url)
                    ),
                )
                enforce(
                    isinstance(self.version, str),
                    "Invalid type for content 'version'. Expected 'str'. Found '{}'.".format(
                        type(self.version)
                    ),
                )
                enforce(
                    isinstance(self.headers, str),
                    "Invalid type for content 'headers'. Expected 'str'. Found '{}'.".format(
                        type(self.headers)
                    ),
                )
                enforce(
                    isinstance(self.body, bytes),
                    "Invalid type for content 'body'. Expected 'bytes'. Found '{}'.".format(
                        type(self.body)
                    ),
                )
            elif self.performative == HttpMessage.Performative.RESPONSE:
                expected_nb_of_contents = 5
                enforce(
                    isinstance(self.version, str),
                    "Invalid type for content 'version'. Expected 'str'. Found '{}'.".format(
                        type(self.version)
                    ),
                )
                enforce(
                    type(self.status_code) is int,
                    "Invalid type for content 'status_code'. Expected 'int'. Found '{}'.".format(
                        type(self.status_code)
                    ),
                )
                enforce(
                    isinstance(self.status_text, str),
                    "Invalid type for content 'status_text'. Expected 'str'. Found '{}'.".format(
                        type(self.status_text)
                    ),
                )
                enforce(
                    isinstance(self.headers, str),
                    "Invalid type for content 'headers'. Expected 'str'. Found '{}'.".format(
                        type(self.headers)
                    ),
                )
                enforce(
                    isinstance(self.body, bytes),
                    "Invalid type for content 'body'. Expected 'bytes'. Found '{}'.".format(
                        type(self.body)
                    ),
                )

            # Check correct content count
            enforce(
                expected_nb_of_contents == actual_nb_of_contents,
                "Incorrect number of contents. Expected {}. Found {}".format(
                    expected_nb_of_contents, actual_nb_of_contents
                ),
            )

            # Light Protocol Rule 3
            if self.message_id == 1:
                enforce(
                    self.target == 0,
                    "Invalid 'target'. Expected 0 (because 'message_id' is 1). Found {}.".format(
                        self.target
                    ),
                )
        except (AEAEnforceError, ValueError, KeyError) as e:
            _default_logger.error(str(e))
            return False

        return True



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Override for hash command."""

import sys
from pathlib import Path
from typing import Optional, cast

import click
from aea.cli.ipfs_hash import (
    IPFSHashOnly,
    PackageConfiguration,
    PackageType,
    check_hashes,
    load_configuration_object,
    update_hashes,
)
from aea.configurations.base import (
    PACKAGE_TYPE_TO_CONFIG_CLASS as _PACKAGE_TYPE_TO_CONFIG_CLASS,
)

from autonomy.configurations.base import Service as ServiceConfig


PACKAGE_TYPE_TO_CONFIG_CLASS = {
    **_PACKAGE_TYPE_TO_CONFIG_CLASS,
    PackageType.SERVICE: ServiceConfig,
}


def load_configuration(
    package_type: PackageType, package_path: Path
) -> PackageConfiguration:
    """
    Load a configuration, knowing the type and the path to the package root.

    :param package_type: the package type.
    :param package_path: the path to the package root.
    :return: the configuration object.
    """
    configuration_obj = load_configuration_object(
        package_type,
        package_path,
        package_type_config_class=PACKAGE_TYPE_TO_CONFIG_CLASS,
    )
    configuration_obj._directory = package_path  # pylint: disable=protected-access
    return cast(PackageConfiguration, configuration_obj)


@click.group(name="hash")
def hash_group() -> None:
    """Hashing utils."""


@hash_group.command(name="all")
@click.option(
    "--packages-dir",
    type=click.Path(exists=True, dir_okay=True, file_okay=False),
    default=Path("packages/"),
)
@click.option("--vendor", type=str)
@click.option("--no-wrap", is_flag=True)
@click.option("--check", is_flag=True)
def generate_all(
    packages_dir: Path,
    vendor: Optional[str],
    no_wrap: bool,
    check: bool,
) -> None:
    """Generate IPFS hashes."""
    packages_dir = Path(packages_dir).absolute()
    if check:
        return_code = check_hashes(
            packages_dir, no_wrap, vendor=vendor, config_loader=load_configuration
        )
    else:
        return_code = update_hashes(
            packages_dir, no_wrap, vendor=vendor, config_loader=load_configuration
        )
    sys.exit(return_code)


@hash_group.command(name="one")
@click.argument("path", type=click.Path(exists=True, file_okay=True, dir_okay=True))
@click.option("--no-wrap", is_flag=True)
def hash_file(path: str, no_wrap: bool) -> None:
    """Hash a single file/directory."""

    click.echo(f"Path : {path}")
    click.echo(f"Hash : {IPFSHashOnly.get(path, wrap=not no_wrap)}")
